import pytest
from lxml import etree

from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.get_data_deserialised import _deserialise_all_resources
from dsp_tools.utils.xml_parsing.get_data_deserialised import _deserialise_one_property
from dsp_tools.utils.xml_parsing.get_data_deserialised import _deserialise_one_resource
from dsp_tools.utils.xml_parsing.get_data_deserialised import _extract_metadata
from dsp_tools.utils.xml_parsing.get_data_deserialised import _get_text_as_string
from dsp_tools.utils.xml_parsing.models.data_deserialised import PropertyObject
from dsp_tools.utils.xml_parsing.models.data_deserialised import ResourceDeserialised
from dsp_tools.utils.xml_parsing.models.data_deserialised import TripleObjectType
from dsp_tools.utils.xml_parsing.models.data_deserialised import TriplePropertyType
from dsp_tools.utils.xml_parsing.models.data_deserialised import ValueInformation
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType


def _get_label_and_type(resource: ResourceDeserialised) -> tuple[PropertyObject, PropertyObject, list[PropertyObject]]:
    lbl = next(x for x in resource.property_objects if x.property_type == TriplePropertyType.RDFS_LABEL)
    rdf_type = next(x for x in resource.property_objects if x.property_type == TriplePropertyType.RDF_TYPE)
    permissions = list(x for x in resource.property_objects if x.property_type == TriplePropertyType.KNORA_PERMISSIONS)
    return lbl, rdf_type, permissions


class TestResource:
    def test_empty(self, resource_empty: etree._Element) -> None:
        res = _deserialise_one_resource(resource_empty)
        assert res.res_id == "one"
        assert len(res.property_objects) == 2
        assert not res.asset_value
        lbl, rdf_type, _ = _get_label_and_type(res)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 0

    def test_migration_metadata(self, resource_with_migration_metadata: etree._Element) -> None:
        res = _deserialise_one_resource(resource_with_migration_metadata)
        assert res.res_id == "one"
        assert len(res.property_objects) == 2
        assert not res.asset_value
        lbl, rdf_type, _ = _get_label_and_type(res)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 0
        assert res.migration_metadata.any()
        assert res.migration_metadata.ark == "ark"
        assert res.migration_metadata.iri == "iri"
        assert str(res.migration_metadata.creation_date) == "2019-01-09T15:45:54.502951Z"

    def test_empty_permissions(self, resource_empty_permissions: etree._Element) -> None:
        res = _deserialise_one_resource(resource_empty_permissions)
        assert res.res_id == "one"
        assert len(res.property_objects) == 3
        assert not res.asset_value
        lbl, rdf_type, perm = _get_label_and_type(res)
        assert len(perm) == 1
        permission = perm.pop(0)
        assert permission.object_value == "open"
        assert permission.object_type == TripleObjectType.STRING
        assert permission.property_type == TriplePropertyType.KNORA_PERMISSIONS
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 0
        assert not res.migration_metadata.any()

    def test_with_props(self, root_resource_with_props: etree._Element) -> None:
        res_list = _deserialise_all_resources(root_resource_with_props).resources
        assert len(res_list) == 1
        res = res_list[0]
        assert res.res_id == "one"
        assert len(res.property_objects) == 2
        assert not res.asset_value
        lbl, rdf_type, _ = _get_label_and_type(res)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 3
        assert not res.migration_metadata.any()

    def test_with_bitstream(self, resource_with_bitstream: etree._Element) -> None:
        result = _deserialise_one_resource(resource_with_bitstream)
        assert len(result.values) == 0
        bitstream = result.asset_value
        assert isinstance(bitstream, ValueInformation)
        assert bitstream.user_facing_prop == f"{KNORA_API_STR}hasAudioFileValue"
        assert bitstream.user_facing_value == "testdata/bitstreams/test.wav"
        assert bitstream.knora_type == KnoraValueType.AUDIO_FILE
        assert not bitstream.value_metadata
        assert not result.migration_metadata.any()

    def test_region(self, root_resource_region: etree._Element) -> None:
        res_list = _deserialise_all_resources(root_resource_region).resources
        res = res_list[0]
        assert res.res_id == "region_1"
        assert len(res.property_objects) == 2
        assert not res.asset_value
        lbl, rdf_type, _ = _get_label_and_type(res)
        assert lbl.object_value == "Region"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://api.knora.org/ontology/knora-api/v2#Region"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 4
        expected_props = {
            f"{KNORA_API_STR}hasColor",
            f"{KNORA_API_STR}isRegionOf",
            f"{KNORA_API_STR}hasGeometry",
            f"{KNORA_API_STR}hasComment",
        }
        actual_props = {x.user_facing_prop for x in res.values}
        assert actual_props == expected_props
        assert not res.migration_metadata.any()

    def test_audio_segment(self, audio_segment: etree._Element) -> None:
        res = _deserialise_one_resource(audio_segment)
        assert res.res_id == "audio_id"
        assert len(res.property_objects) == 2
        assert not res.asset_value
        lbl, rdf_type, _ = _get_label_and_type(res)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://api.knora.org/ontology/knora-api/v2#AudioSegment"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 2
        propname_to_info = {x.user_facing_prop: x for x in res.values}
        segment_of = propname_to_info.pop(f"{KNORA_API_STR}isAudioSegmentOf")
        assert segment_of.user_facing_value == "is_segment_of_id"
        assert segment_of.knora_type == KnoraValueType.LINK_VALUE
        assert len(segment_of.value_metadata) == 1
        assert segment_of.value_metadata[0].property_type == TriplePropertyType.KNORA_COMMENT_ON_VALUE
        assert segment_of.value_metadata[0].object_value == "Comment"
        assert segment_of.value_metadata[0].object_type == TripleObjectType.STRING
        segment_bounds = propname_to_info.pop(f"{KNORA_API_STR}hasSegmentBounds")
        assert not segment_bounds.user_facing_value
        assert segment_bounds.knora_type == KnoraValueType.INTERVAL_VALUE
        assert len(segment_bounds.value_metadata) == 3
        seg_bounds_prop_objects = {x.property_type: x for x in segment_bounds.value_metadata}
        permission = seg_bounds_prop_objects.pop(TriplePropertyType.KNORA_PERMISSIONS)
        assert permission.object_value == "open"
        assert permission.object_type == TripleObjectType.STRING
        start_bound = seg_bounds_prop_objects.pop(TriplePropertyType.KNORA_INTERVAL_START)
        assert start_bound.object_value == "0.5"
        assert start_bound.object_type == TripleObjectType.DECIMAL
        end_bound = seg_bounds_prop_objects.pop(TriplePropertyType.KNORA_INTERVAL_END)
        assert end_bound.object_value == "7"
        assert end_bound.object_type == TripleObjectType.DECIMAL
        assert not res.migration_metadata.any()

    def test_video_segment(self, video_segment: etree._Element) -> None:
        res = _deserialise_one_resource(video_segment)
        assert res.res_id == "video_id"
        assert len(res.property_objects) == 2
        assert not res.asset_value
        lbl, rdf_type, _ = _get_label_and_type(res)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == "http://api.knora.org/ontology/knora-api/v2#VideoSegment"
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(res.values) == 7
        propname_to_info = {x.user_facing_prop: x for x in res.values}
        segment_bounds = propname_to_info.pop(f"{KNORA_API_STR}hasSegmentBounds")
        assert not segment_bounds.user_facing_value
        assert segment_bounds.knora_type == KnoraValueType.INTERVAL_VALUE
        assert len(segment_bounds.value_metadata) == 2
        seg_bounds_prop_objects = {x.property_type: x for x in segment_bounds.value_metadata}
        start_bound = seg_bounds_prop_objects.pop(TriplePropertyType.KNORA_INTERVAL_START)
        assert start_bound.object_value == "0.1"
        assert start_bound.object_type == TripleObjectType.DECIMAL
        end_bound = seg_bounds_prop_objects.pop(TriplePropertyType.KNORA_INTERVAL_END)
        assert end_bound.object_value == "0.234"
        assert end_bound.object_type == TripleObjectType.DECIMAL

        other_props = {
            f"{KNORA_API_STR}isVideoSegmentOf": ("is_segment_of_id", KnoraValueType.LINK_VALUE),
            f"{KNORA_API_STR}hasTitle": ("Title", KnoraValueType.SIMPLETEXT_VALUE),
            f"{KNORA_API_STR}hasComment": ("Comment", KnoraValueType.RICHTEXT_VALUE),
            f"{KNORA_API_STR}hasKeyword": ("Keyword", KnoraValueType.SIMPLETEXT_VALUE),
            f"{KNORA_API_STR}relatesTo": ("relates_to_id", KnoraValueType.LINK_VALUE),
            f"{KNORA_API_STR}hasDescription": ("Description", KnoraValueType.RICHTEXT_VALUE),
        }
        for prop_name, value_info in propname_to_info.items():
            expected = other_props[prop_name]
            assert value_info.user_facing_value == expected[0]
            assert value_info.knora_type == expected[1]
            assert len(value_info.value_metadata) == 0
        assert not res.migration_metadata.any()


class TestBooleanValue:
    def test_corr(self, boolean_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(boolean_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean"
        assert res.user_facing_value == "true"
        assert res.knora_type == KnoraValueType.BOOLEAN_VALUE
        assert not res.value_metadata


class TestColorValue:
    def test_corr(self, color_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(color_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testColor"
        assert res.user_facing_value == "#00ff00"
        assert res.knora_type == KnoraValueType.COLOR_VALUE
        assert not res.value_metadata

    def test_several(self, color_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(color_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testColor"
        assert res[0].user_facing_value == "#00ff00"
        assert res[1].user_facing_value == "#00ff11"
        assert res[0].knora_type == KnoraValueType.COLOR_VALUE
        assert res[1].knora_type == KnoraValueType.COLOR_VALUE


class TestDateValue:
    def test_corr(self, date_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(date_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1"
        assert res.user_facing_value == "JULIAN:BCE:0700:BCE:0600"
        assert res.knora_type == KnoraValueType.DATE_VALUE
        assert not res.value_metadata

    def test_several(self, date_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(date_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1"
        assert res[0].user_facing_value == "JULIAN:BCE:0700:BCE:0600"
        assert res[1].user_facing_value == "ISLAMIC:0600:0700"
        assert res[0].knora_type == KnoraValueType.DATE_VALUE
        assert res[1].knora_type == KnoraValueType.DATE_VALUE


class TestDecimalValue:
    def test_corr(self, decimal_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(decimal_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText"
        assert res.user_facing_value == "2.71"
        assert res.knora_type == KnoraValueType.DECIMAL_VALUE
        assert not res.value_metadata

    def test_several(self, decimal_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(decimal_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText"
        assert res[0].user_facing_value == "1.0"
        assert res[1].user_facing_value == "2.0"
        assert res[0].knora_type == KnoraValueType.DECIMAL_VALUE
        assert res[1].knora_type == KnoraValueType.DECIMAL_VALUE


class TestGeonameValue:
    def test_corr(self, geoname_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(geoname_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname"
        assert res.user_facing_value == "1111111"
        assert res.knora_type == KnoraValueType.GEONAME_VALUE
        assert not res.value_metadata

    def test_several(self, geoname_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(geoname_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname"
        assert res[0].user_facing_value == "1111111"
        assert res[1].user_facing_value == "2222222"
        assert res[0].knora_type == KnoraValueType.GEONAME_VALUE
        assert res[1].knora_type == KnoraValueType.GEONAME_VALUE


class TestGeomValue:
    def test_corr(self, geometry_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(geometry_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == f"{KNORA_API_STR}hasGeometry"
        assert res.user_facing_value is not None
        assert res.knora_type == KnoraValueType.GEOM_VALUE
        assert not res.value_metadata

    def test_wrong(self, geometry_value_wrong: etree._Element) -> None:
        res_list = _deserialise_one_property(geometry_value_wrong)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == f"{KNORA_API_STR}hasGeometry"
        assert not res.user_facing_value
        assert res.knora_type == KnoraValueType.GEOM_VALUE
        assert not res.value_metadata


class TestIntValue:
    def test_corr(self, integer_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(integer_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText"
        assert res.user_facing_value == "1"
        assert res.knora_type == KnoraValueType.INT_VALUE
        assert not res.value_metadata

    def test_several(self, integer_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(integer_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText"
        assert res[0].user_facing_value == "1"
        assert res[1].user_facing_value == "2"
        assert res[0].knora_type == KnoraValueType.INT_VALUE
        assert res[1].knora_type == KnoraValueType.INT_VALUE


class TestListValue:
    def test_corr(self, list_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(list_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp"
        assert res.user_facing_value == "firstList / n1"
        assert res.knora_type == KnoraValueType.LIST_VALUE
        assert not res.value_metadata

    def test_several(self, list_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(list_value_corr_several)
        assert len(res) == 2
        one = res[0]
        assert one.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp"
        assert res[0].user_facing_value == "firstList / n1"
        assert res[1].user_facing_value == "firstList / n2"
        assert res[0].knora_type == KnoraValueType.LIST_VALUE
        assert res[1].knora_type == KnoraValueType.LIST_VALUE


class TestSimpleTextValue:
    def test_corr(self, text_simpletext_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(text_simpletext_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea"
        assert res.user_facing_value == "Text"
        assert res.knora_type == KnoraValueType.SIMPLETEXT_VALUE
        assert not res.value_metadata

    def test_several(self, text_simpletext_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(text_simpletext_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText"
        assert res[0].user_facing_value == "Text 1"
        assert res[1].user_facing_value == "Text 2"
        assert res[0].knora_type == KnoraValueType.SIMPLETEXT_VALUE
        assert res[1].knora_type == KnoraValueType.SIMPLETEXT_VALUE

    def test_wrong(self, text_simpletext_value_wrong: etree._Element) -> None:
        res_list = _deserialise_one_property(text_simpletext_value_wrong)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText"
        assert not res.user_facing_value
        assert res.knora_type == KnoraValueType.SIMPLETEXT_VALUE
        assert not res.value_metadata


class TestRichtextValue:
    def test_corr(self, text_richtext_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(text_richtext_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext"
        assert res.user_facing_value == "<p>Text</p>"
        assert res.knora_type == KnoraValueType.RICHTEXT_VALUE
        assert not res.value_metadata

    def test_several(self, text_richtext_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(text_richtext_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext"
        assert res[0].user_facing_value == "Text 1"
        assert res[1].user_facing_value == "Text 2"
        assert res[0].knora_type == KnoraValueType.RICHTEXT_VALUE
        assert res[1].knora_type == KnoraValueType.RICHTEXT_VALUE


class TestTimeValue:
    def test_corr(self, time_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(time_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue"
        assert res.user_facing_value == "2019-10-23T13:45:12.01-14:00"
        assert res.knora_type == KnoraValueType.TIME_VALUE
        assert not res.value_metadata

    def test_several(self, time_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(time_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue"
        assert res[0].user_facing_value == "2019-10-23T13:45:12.01-14:00"
        assert res[1].user_facing_value == "2019-10-23T13:45:12.01-08:00"
        assert res[0].knora_type == KnoraValueType.TIME_VALUE
        assert res[1].knora_type == KnoraValueType.TIME_VALUE


class TestUriValue:
    def test_corr(self, uri_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(uri_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue"
        assert res.user_facing_value == "https://dasch.swiss"
        assert res.knora_type == KnoraValueType.URI_VALUE
        assert not res.value_metadata

    def test_several(self, uri_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(uri_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue"
        assert res[0].user_facing_value == "https://dasch.swiss"
        assert res[1].user_facing_value == "https://app.dasch.swiss"
        assert res[0].knora_type == KnoraValueType.URI_VALUE
        assert res[1].knora_type == KnoraValueType.URI_VALUE


class TestLinkValue:
    def test_corr(self, resptr_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(resptr_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert res.user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo"
        assert res.user_facing_value == "id_1"
        assert res.knora_type == KnoraValueType.LINK_VALUE
        assert not res.value_metadata

    def test_several(self, resptr_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(resptr_value_corr_several)
        assert len(res) == 2
        assert res[0].user_facing_prop == "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo"
        assert res[0].user_facing_value == "id_1"
        assert res[1].user_facing_value == "id_2"
        assert res[0].knora_type == KnoraValueType.LINK_VALUE
        assert res[1].knora_type == KnoraValueType.LINK_VALUE


class TestFileValue:
    def test_iiif(self, iiif_with_spaces: etree._Element) -> None:
        result = _deserialise_one_property(iiif_with_spaces)
        assert len(result) == 1
        res = result.pop()
        assert res.user_facing_prop == f"{KNORA_API_STR}hasStillImageFileValue"
        assert res.user_facing_value == "https://iiif.uri/full.jpg"
        assert res.knora_type == KnoraValueType.STILL_IMAGE_IIIF
        assert not res.value_metadata


class TestExtractMetadata:
    def test_none(self, boolean_value_no_attrib: etree._Element) -> None:
        res = _extract_metadata(boolean_value_no_attrib)
        assert not res

    def test_comment(self, boolean_value_comment: etree._Element) -> None:
        res = _extract_metadata(boolean_value_comment)
        assert len(res) == 1
        val = res.pop(0)
        assert val.property_type == TriplePropertyType.KNORA_COMMENT_ON_VALUE
        assert val.object_value == "Comment on Value"
        assert val.object_type == TripleObjectType.STRING

    def test_permissions(self, boolean_value_permissions: etree._Element) -> None:
        res = _extract_metadata(boolean_value_permissions)
        assert len(res) == 1
        permission = res.pop(0)
        assert permission.property_type == TriplePropertyType.KNORA_PERMISSIONS
        assert permission.object_value == "open"
        assert permission.object_type == TripleObjectType.STRING

    def test_iiif_with_legal_info(self, iiif_with_legal_info: etree._Element) -> None:
        result = _extract_metadata(iiif_with_legal_info)
        assert len(result) == 3
        license_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_LICENSE)
        assert license_res.object_value == "license_iri"
        assert license_res.object_type == TripleObjectType.IRI
        author_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_AUTHORSHIP)
        assert author_res.object_value == "auth"
        assert author_res.object_type == TripleObjectType.STRING
        copyright_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_COPYRIGHT_HOLDER)
        assert copyright_res.object_value == "copy"
        assert copyright_res.object_type == TripleObjectType.STRING


@pytest.mark.parametrize(
    ("input_str", "expected"),
    [
        ('<text encoding="xml"><p>Text</p></text>', "<p>Text</p>"),
        ('<text encoding="xml"><text>Text</text></text>', "<text>Text</text>"),
        ('<text encoding="xml"><p><p>Text second word</p>Tail</p></text>', "<p><p>Text second word</p>Tail</p>"),
        ('<text encoding="xml"> </text>', " "),
        ('<text encoding="xml"></text>', None),
        ('<text encoding="xml"><br/>Text</text>', "<br/>Text"),
        ('<text encoding="xml">Text<br/>  </text>', "Text<br/>"),
        ('<text encoding="xml">  Text<br/>Text<br/>Text</text>', "Text<br/>Text<br/>Text"),
    ],
)
def test_get_text_as_string(input_str: str, expected: str) -> None:
    val = etree.fromstring(input_str)
    result = _get_text_as_string(val)
    assert result == expected


if __name__ == "__main__":
    pytest.main([__file__])



"""
This module implements the handling (CRUD) of list nodes and adds some function to read whole lists.

CREATE:
    * Instantiate a new object of the class ListNode
    * Call the ``create`` method on the instance

READ:
    * Instantiate a new object with ``iri`` (IRI of listnode)
    * Call the ``read`` method on the instance
    * Access information about the instance
"""

from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.project.legacy_models.model import Model
from dsp_tools.commands.project.legacy_models.project import Project
from dsp_tools.error.exceptions import BaseError
from dsp_tools.legacy_models.langstring import LangString


class ListNode(Model):
    """
    This class represents a list node

    Attributes
    ----------

    con : Connection
        A Connection instance to a DSP server (for some operation a login has to be performed with valid credentials)

    iri : str
        IRI of the list node [readonly, cannot be modified after creation of instance]

    project : str
        IRI of project. Only used for the creation of a new list (root node) [write].

    label : LangString
        A LangString instance with language dependent labels. Setting this attribute overwrites all entries
        with the new ones.
        At least one label is required [read/write].

    comments : LangString
        A LangString instance with language dependent comments. Setting this attributes overwrites all entries
        with the new ones.

    name : str
        A unique name for the ListNode (unique inside this list) [read/write].

    parent : IRI | ListNode
        Is required and allowed only for the CREATE operation. Otherwise use the
        "children" attribute [write].

    isRootNode : bool
        Is True if the ListNode is the root node of a list. Cannot be set [read].

    children : list[ListNode]
        Contains a list of child nodes. This attribute is only available for nodes that have been read by the
        method "getAllNodes()" [read].

    rootNodeIri : str
        IRI of the root node. This attribute is only available for nodes that have been read by the
        method "getAllNodes()" [read].

    Methods
    -------

    create : ListNode information object
        Creates a new list (node) and returns the information from the list (node). Use it to create new lists
        or append new ListNodes to an existing list.

    read : ListNode information object
        Returns information about a single list node

    getAllNodes : ListNode
        Get all nodes of a list. The IRI of the root node has to be supplied.

    getAllLists [static]:
        Returns all lists of a project.

    print : None
        Prints the ListNode information to stdout (not recursive)

    """

    ROUTE = "/admin/lists"
    ROUTE_SLASH = ROUTE + "/"

    _id: Optional[str]
    _project: Optional[str]
    _label: Optional[LangString]
    _comments: Optional[LangString]
    _name: Optional[str]
    _parent: Optional[str]
    _isRootNode: bool
    _children: Optional[list[ListNode]]
    _rootNodeIri: Optional[str]

    def __init__(
        self,
        con: Connection,
        iri: Optional[str] = None,
        project: Optional[Union[Project, str]] = None,
        label: Optional[LangString] = None,
        comments: Optional[LangString] = None,
        name: Optional[str] = None,
        parent: Optional[Union[ListNode, str]] = None,
        isRootNode: bool = False,
        children: Optional[list[ListNode]] = None,
        rootNodeIri: Optional[str] = None,
    ) -> None:
        """
        This is the constructor for the ListNode object. For

        CREATE:
            * The "con" and at least one "label" are required
        READ:
            * The "con" and "iri" attributes are required

        Args:
            con: A valid Connection instance with a user logged in that has the appropriate permissions
            iri: IRI of the project [readonly, cannot be modified after creation of instance]
            project: IRI of project. Only used for the creation of a new list (root node) [write].
            label: LangString with lang dependent labels. At least one label is required [read/write].
            comments:  A LangString instance with language dependent comments.
            name: A unique name for the ListNode (unique regarding the whole list) [read/write].
            parent: Required and allowed only for the CREATE operation. Otherwise use the "children" attribute [write].
            isRootNode: Is True if the ListNode is a root node of a list Cannot be set [read].
            children: list of children nodes. Only available for nodes that have been read by getAllNodes()
            rootNodeIri: IRI of the root node. Only available for nodes that have been read by the method getAllNodes()
        """

        super().__init__(con)

        self._project = project.iri if isinstance(project, Project) else str(project) if project else None
        self._id = iri
        self._label = LangString(label)
        self._comments = LangString(comments) if comments else None
        self._name = name
        if parent and isinstance(parent, ListNode):
            self._parent = parent.iri
        else:
            self._parent = str(parent) if parent else None
        self._isRootNode = isRootNode
        if children:
            if isinstance(children, list) and len(children) > 0 and isinstance(children[0], ListNode):
                self._children = children
            else:
                raise BaseError("ERROR Children must be list of ListNodes!")
        else:
            self._children = None
        if not isinstance(rootNodeIri, str) and rootNodeIri:
            raise BaseError("ERROR rootNodeIri must be of type string")
        self._rootNodeIri = rootNodeIri

    @property
    def iri(self) -> Optional[str]:
        return self._id

    @property
    def project(self) -> Optional[str]:
        return self._project

    @project.setter
    def project(self, value: str) -> None:
        if self._project:
            raise BaseError("Project id cannot be modified!")
        else:
            self._project = value

    @property
    def label(self) -> LangString:
        return self._label or LangString({})

    @label.setter
    def label(self, value: Optional[Union[LangString, str]]) -> None:
        self._label = LangString(value)
        self._changed.add("label")

    @property
    def comments(self) -> LangString:
        return self._comments or LangString({})

    @comments.setter
    def comments(self, value: Optional[Union[LangString, str]]) -> None:
        self._comments = LangString(value)
        self._changed.add("comments")

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        self._changed.add("name")

    @property
    def parent(self) -> Optional[str]:
        return self._parent if self._parent else None

    @parent.setter
    def parent(self, value: Union[str, ListNode]) -> None:
        if isinstance(value, ListNode):
            self._parent = value.iri
        else:
            self._parent = value

    @property
    def isRootNode(self) -> Optional[bool]:
        return self._isRootNode

    @property
    def children(self) -> list[ListNode]:
        return self._children or []

    @children.setter
    def children(self, value: list[ListNode]) -> None:
        self._children = value

    @staticmethod
    def __getChildren(
        con: Connection,
        parent_iri: str,
        project_iri: str,
        children: list[Any],
    ) -> Optional[list[ListNode]]:
        """
        Internal method! Should not be used directly!

        Static method gets a recursive List of children nodes

        :param con: Valid Connection instance
        :param children: json object of children
        :return: List of ListNode instances
        """
        if children:
            child_nodes: list[Any] = []
            for child in children:
                if "parentNodeIri" not in child:
                    child["parentNodeIri"] = parent_iri
                if "projectIri" not in child:
                    child["projectIri"] = project_iri
                child_node = ListNode.fromJsonObj(con, child)
                child_nodes.append(child_node)
            return child_nodes
        else:
            return None

    @property
    def rootNodeIri(self) -> Optional[str]:
        return self._rootNodeIri

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> ListNode:
        """
        Internal method! Should not be used directly!

        This method is used to create a ListNode instance from the JSON data returned by DSP

        :param con: Connection instance
        :param json_obj: JSON data returned by DSP as python3 object
        :return: ListNode instance
        """

        iri = json_obj.get("id")
        if iri is None:
            raise BaseError("ListNode id is missing")
        project = json_obj.get("projectIri")
        label = LangString.fromJsonObj(json_obj.get("labels"))
        comments = LangString.fromJsonObj(json_obj.get("comments"))
        if json_obj.get("name"):
            name = json_obj["name"]
        else:
            name = iri.rsplit("/", 1)[-1]
        parent = json_obj.get("parentNodeIri")
        isRootNode = json_obj.get("isRootNode")

        child_info = json_obj.get("children")
        children = None
        if child_info:
            children = ListNode.__getChildren(con=con, parent_iri=iri, project_iri=project, children=child_info)
        rootNodeIri = json_obj.get("hasRootNode")

        return cls(
            con=con,
            iri=iri,
            project=project,
            label=label,
            comments=comments,
            name=name,
            parent=parent,
            isRootNode=isRootNode,
            children=children,
            rootNodeIri=rootNodeIri,
        )

    def create(self) -> ListNode:
        """
        Create a new List

        :return: JSON-object from DSP-API
        """
        jsonobj = self._toJsonObj_create()
        if self._parent:
            result = self._con.post(ListNode.ROUTE_SLASH + quote_plus(self._parent), jsonobj)
            return ListNode.fromJsonObj(self._con, result["nodeinfo"])
        else:
            result = self._con.post(ListNode.ROUTE, jsonobj)
            return ListNode.fromJsonObj(self._con, result["list"]["listinfo"])

    def _toJsonObj_create(self):
        tmp = {}
        if self._project is None:
            raise BaseError("There must be a project id given!")
        tmp["projectIri"] = self._project
        if self._label.isEmpty():
            raise BaseError("There must be a valid ListNode label!")
        tmp["labels"] = self._label.toJsonObj()
        if self._comments:
            tmp["comments"] = self._comments.toJsonObj()
        if self._name:
            tmp["name"] = self._name
        if self._parent:
            tmp["parentNodeIri"] = self._parent
        return tmp

    def read(self) -> Any:
        """
        Read a project from DSP-API

        :return: JSON-object from DSP-API
        """

        result = self._con.get(ListNode.ROUTE_SLASH + "nodes/" + quote_plus(self._id))
        if result.get("nodeinfo"):
            return self.fromJsonObj(self._con, result["nodeinfo"])
        elif result.get("listinfo"):
            return self.fromJsonObj(self._con, result["listinfo"])
        else:
            return None

    def getAllNodes(self) -> ListNode:
        """
        Get all nodes of the list. Must be called from a ListNode instance that has at least set the
        list iri!

        :return: Root node of list with recursive ListNodes ("children"-attributes)
        """

        result = self._con.get(ListNode.ROUTE_SLASH + quote_plus(self._id))
        if "list" not in result:
            raise BaseError("Request got no list!")
        if "listinfo" not in result["list"]:
            raise BaseError("Request got no proper list information!")
        root = ListNode.fromJsonObj(self._con, result["list"]["listinfo"])
        if "children" in result["list"]:
            root.children = ListNode.__getChildren(
                con=self._con, parent_iri=root.iri, project_iri=root.project, children=result["list"]["children"]
            )
        return root

    @staticmethod
    def getAllLists(con: Connection, project_iri: Optional[str] = None) -> list[ListNode]:
        """
        Get all lists. If a project IRI is given, it returns the lists of the specified project

        :param con: Connection instance
        :param project_iri: Iri/id of project
        :return: list of ListNodes
        """
        if project_iri is None:
            result = con.get(ListNode.ROUTE)
        else:
            result = con.get(ListNode.ROUTE + "?projectIri=" + quote_plus(project_iri))
        if "lists" not in result:
            raise BaseError("Request got no lists!")
        return list(map(lambda a: ListNode.fromJsonObj(con, a), result["lists"]))

    def _createDefinitionFileObj(self, children: list[ListNode]) -> list[dict[str, Any]]:
        """
        Create an object that corresponds to the syntax of the input to "create_onto".
        Node: This method must be used only internally (for recursion)!!

        :param children: List of children nodes
        :return: A python object that can be jsonfied to correspond to the syntax of the input to "create_onto".
        """
        listnodeobjs = []
        for listnode in children:
            listnodeobj = {
                "name": listnode.name,
                "labels": listnode.label.createDefinitionFileObj(),
            }
            if not listnode.comments.isEmpty():
                listnodeobj["comments"] = listnode.comments.createDefinitionFileObj()
            if listnode.children:
                listnodeobj["nodes"] = self._createDefinitionFileObj(listnode.children)
            listnodeobjs.append(listnodeobj)
        return listnodeobjs

    def createDefinitionFileObj(self) -> dict[str, Any]:
        """
        Create an object that corresponds to the syntax of the input to "create_onto".
        :return: A python object that can be jsonfied to correspond to the syntax of the input to "create_onto".
        """
        listnode = {
            "name": self._name,
            "labels": self._label.createDefinitionFileObj(),
        }
        if not self._comments.isEmpty():
            listnode["comments"] = self._comments.createDefinitionFileObj()
        if self._children:
            listnode["nodes"] = self._createDefinitionFileObj(self._children)
        return listnode



from dataclasses import dataclass
from typing import cast

from rdflib import Graph

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.xmlupload.make_rdf_graph.jsonld_utils import serialise_jsonld_for_resource


@dataclass(frozen=True)
class ResourceCreateClient:
    """client class that creates resources on a DSP server."""

    con: Connection

    def create_resource(
        self,
        graph: Graph,
        resource_has_bitstream: bool,
    ) -> str:
        """Creates a resource on the DSP server, and returns its IRI"""
        res_dict = serialise_jsonld_for_resource(graph)
        headers = {"X-Asset-Ingested": "true"} if resource_has_bitstream else None
        res = self.con.post(route="/v2/resources", data=res_dict, headers=headers)
        return cast(str, res["@id"])



from dataclasses import dataclass
from typing import Any

import pytest

from dsp_tools.commands.xmlupload.prepare_xml_input.list_client import List
from dsp_tools.commands.xmlupload.prepare_xml_input.list_client import ListClientLive
from dsp_tools.commands.xmlupload.prepare_xml_input.list_client import ListNode
from dsp_tools.commands.xmlupload.prepare_xml_input.list_client import _get_list_from_server
from dsp_tools.commands.xmlupload.prepare_xml_input.list_client import _get_list_iris_from_server
from test.integration.commands.xmlupload.connection_mock import ConnectionMockBase


@dataclass
class ConnectionMock(ConnectionMockBase):
    get_responses: list[dict[str, Any]]

    def get(
        self,
        route: str,  # noqa: ARG002 (unused-method-argument)
        headers: dict[str, str] | None = None,  # noqa: ARG002 (unused-method-argument)
    ) -> dict[str, Any]:
        return self.get_responses.pop(0)


class TestGetListNodeLookup:
    def test_list_root_only(self) -> None:
        list_iris = {"lists": [{"id": "http://www.example.org/lists#a"}]}
        list_a = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "name": "list-a",
                },
                "children": [],
            }
        }
        con = ConnectionMock([list_iris, list_a])
        list_client = ListClientLive(con, "")
        lookup = list_client.get_list_node_id_to_iri_lookup()
        expected = {("list-a", "list-a"): "http://www.example.org/lists#a"}
        assert lookup == expected

    def test_list_with_children(self) -> None:
        list_iris = {"lists": [{"id": "http://www.example.org/lists#a"}]}
        list_a = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "name": "list-a",
                },
                "children": [
                    {
                        "id": "http://www.example.org/lists#a1",
                        "name": "node-a1",
                    },
                    {
                        "id": "http://www.example.org/lists#a2",
                        "name": "node-a2",
                    },
                ],
            }
        }
        con = ConnectionMock([list_iris, list_a])
        list_client = ListClientLive(con, "")
        lookup = list_client.get_list_node_id_to_iri_lookup()
        expected = {
            ("list-a", "list-a"): "http://www.example.org/lists#a",
            ("list-a", "node-a1"): "http://www.example.org/lists#a1",
            ("list-a", "node-a2"): "http://www.example.org/lists#a2",
        }
        assert lookup == expected

    def test_multiple_lists(self) -> None:
        list_iris = {
            "lists": [
                {"id": "http://www.example.org/lists#a"},
                {"id": "http://www.example.org/lists#b"},
            ]
        }
        list_a = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "name": "list-a",
                },
                "children": [
                    {
                        "id": "http://www.example.org/lists#a1",
                        "name": "node-a1",
                    },
                ],
            }
        }
        list_b = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#b",
                    "name": "list-b",
                },
                "children": [
                    {
                        "id": "http://www.example.org/lists#b1",
                        "name": "node-b1",
                    },
                ],
            }
        }
        con = ConnectionMock([list_iris, list_a, list_b])
        list_client = ListClientLive(con, "")
        lookup = list_client.get_list_node_id_to_iri_lookup()
        expected = {
            ("list-a", "list-a"): "http://www.example.org/lists#a",
            ("list-a", "node-a1"): "http://www.example.org/lists#a1",
            ("list-b", "list-b"): "http://www.example.org/lists#b",
            ("list-b", "node-b1"): "http://www.example.org/lists#b1",
        }
        assert lookup == expected


class TestGetListFromServer:
    def test_no_name_no_children(self) -> None:
        list_response = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "labels": [{"value": "some label"}],
                },
                "children": [],
            }
        }
        con = ConnectionMock([list_response])
        result = _get_list_from_server(con, "")
        expected_node = ListNode(node_iri="http://www.example.org/lists#a", node_name="some label")
        expected = List(
            root_iri="http://www.example.org/lists#a",
            list_name="some label",
            nodes=[expected_node],
        )
        assert result == expected

    def test_name_no_children(self) -> None:
        list_response = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "name": "list-a",
                },
                "children": [],
            }
        }
        con = ConnectionMock([list_response])
        result = _get_list_from_server(con, "")
        expected_node = ListNode(node_iri="http://www.example.org/lists#a", node_name="list-a")
        expected = List(
            root_iri="http://www.example.org/lists#a",
            list_name="list-a",
            nodes=[expected_node],
        )
        assert result == expected

    def test_with_children(self) -> None:
        list_response = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "name": "list-a",
                },
                "children": [
                    {
                        "id": "http://www.example.org/lists#a1",
                        "name": "node-a1",
                    },
                    {
                        "id": "http://www.example.org/lists#a2",
                        "name": "node-a2",
                    },
                ],
            }
        }
        con = ConnectionMock([list_response])
        result = _get_list_from_server(con, "")
        expected_node_root = ListNode(node_iri="http://www.example.org/lists#a", node_name="list-a")
        expected_node_1 = ListNode(node_iri="http://www.example.org/lists#a1", node_name="node-a1")
        expected_node_2 = ListNode(node_iri="http://www.example.org/lists#a2", node_name="node-a2")
        expected = List(
            root_iri="http://www.example.org/lists#a",
            list_name="list-a",
            nodes=[expected_node_root, expected_node_1, expected_node_2],
        )
        assert result == expected

    def test_multiple_levels(self) -> None:
        list_response = {
            "list": {
                "listinfo": {
                    "id": "http://www.example.org/lists#a",
                    "name": "list-a",
                },
                "children": [
                    {
                        "id": "http://www.example.org/lists#a1",
                        "name": "node-a1",
                        "children": [
                            {
                                "id": "http://www.example.org/lists#a1.1",
                                "name": "node-a1.1",
                            },
                            {
                                "id": "http://www.example.org/lists#a1.2",
                                "name": "node-a1.2",
                            },
                        ],
                    }
                ],
            }
        }
        con = ConnectionMock([list_response])
        result = _get_list_from_server(con, "")
        expected_node_root = ListNode(node_iri="http://www.example.org/lists#a", node_name="list-a")
        expected_node_1 = ListNode(node_iri="http://www.example.org/lists#a1", node_name="node-a1")
        expected_node_1_1 = ListNode(node_iri="http://www.example.org/lists#a1.1", node_name="node-a1.1")
        expected_node_1_2 = ListNode(node_iri="http://www.example.org/lists#a1.2", node_name="node-a1.2")
        expected = List(
            root_iri="http://www.example.org/lists#a",
            list_name="list-a",
            nodes=[expected_node_root, expected_node_1, expected_node_1_1, expected_node_1_2],
        )
        assert result == expected


def test_get_list_iris_from_server() -> None:
    list_iris = {
        "lists": [
            {"id": "http://www.example.org/lists#a"},
            {"id": "http://www.example.org/lists#b"},
        ]
    }
    con = ConnectionMock([list_iris])
    result = _get_list_iris_from_server(con, "")
    expected = ["http://www.example.org/lists#a", "http://www.example.org/lists#b"]
    assert result == expected


if __name__ == "__main__":
    pytest.main([__file__])



from pathlib import Path

import pandas as pd
import pytest

from dsp_tools.commands.ingest_xmlupload.upload_files.filechecker import check_files
from dsp_tools.commands.ingest_xmlupload.upload_files.input_error import FileProblems


def test_check_files_success() -> None:
    supported_and_existing = {
        Path("testdata/bitstreams/test.txt"),
        Path("testdata/bitstreams/test.pdf"),
        Path("testdata/bitstreams/test,with,multiple,commas.png"),
        Path("testdata/bitstreams/test with whitespaces.png"),
        Path("testdata/bitstreams/test with uppercase extension.JPG"),
        Path("testdata/bitstreams/test.with.multiple.dots.png"),
    }
    res = check_files(supported_and_existing)
    assert not res


def test_check_files_unsupported() -> None:
    unsupported_paths = {
        Path("testdata/invalid-testdata/bitstreams/test.gif"),
        Path("testdata/invalid-testdata/bitstreams/test.rtf"),
    }
    res = check_files(unsupported_paths)
    expected = FileProblems([], list(unsupported_paths))
    assert res == expected


def test_check_files_non_existing() -> None:
    inexisting_paths = {Path("foo/bar.txt"), Path("egg/spam.pdf")}
    res = check_files(inexisting_paths)
    expected = FileProblems(list(inexisting_paths), [])
    assert res == expected


def test_check_files_non_existing_and_unsupported() -> None:
    inexisting_unsupported_paths = {Path("foo/bar.baz"), Path("egg/spam.ham")}
    res = check_files(inexisting_unsupported_paths)
    expected = FileProblems([], list(inexisting_unsupported_paths))
    assert res == expected


def test_check_files_mixed_unsupported() -> None:
    mixed_unsupported_paths = {Path("foo/bar.baz"), Path("testdata/invalid-testdata/bitstreams/test.gif")}
    res = check_files(mixed_unsupported_paths)
    expected = FileProblems([], list(mixed_unsupported_paths))
    assert res == expected


class TestFileProblems:
    def test_value_error(self) -> None:
        with pytest.raises(ValueError):  # noqa: PT011 (too broad exception clause)
            FileProblems([], [])

    def test_execute_error_protocol_nonexisting(self) -> None:
        file_problems = FileProblems([Path("foo/bar.txt"), Path("spam/eggs.pdf")], [])
        expected = (
            "Some files referenced in the <bitstream> tags of your XML file cannot be uploaded to the server.\n"
            "\n"
            "The following files don't exist on your computer:\n"
            " - foo/bar.txt\n"
            " - spam/eggs.pdf"
        )
        res = file_problems.execute_error_protocol()
        assert res == expected

    def test_execute_error_protocol_unsupported(self) -> None:
        file_problems = FileProblems([], [Path("testdata/bitstreams/test.txt"), Path("testdata/bitstreams/test.pdf")])
        expected = (
            "Some files referenced in the <bitstream> tags of your XML file cannot be uploaded to the server.\n"
            "\n"
            "The following files have unsupported extensions:\n"
            " - testdata/bitstreams/test.txt\n"
            " - testdata/bitstreams/test.pdf"
        )
        res = file_problems.execute_error_protocol()
        assert res == expected

    def test_execute_error_protocol_inexisting_and_unsupported(self) -> None:
        file_problems = FileProblems(
            [Path("foo/bar.txt"), Path("spam/eggs.pdf")],
            [Path("testdata/bitstreams/test.txt"), Path("testdata/bitstreams/test.pdf")],
        )
        expected = (
            "Some files referenced in the <bitstream> tags of your XML file cannot be uploaded to the server.\n"
            "\n"
            "The following files don't exist on your computer:\n"
            " - foo/bar.txt\n"
            " - spam/eggs.pdf\n"
            "\n"
            "The following files have unsupported extensions:\n"
            " - testdata/bitstreams/test.txt\n"
            " - testdata/bitstreams/test.pdf"
        )
        res = file_problems.execute_error_protocol()
        assert res == expected

    def test_execute_error_protocol_with_file_saving(self) -> None:
        file_problems = FileProblems(
            [Path("foo/bar.txt"), Path("spam/eggs.pdf")], [Path("testdata/bitstreams/test.pdf")], maximum_prints=2
        )
        expected = (
            "Some files referenced in the <bitstream> tags of your XML file cannot be uploaded to the server. "
            "The full list of files with problems has been saved to 'file_problems.csv'."
        )
        res = file_problems.execute_error_protocol()
        assert res == expected

        output_file = Path("file_problems.csv")
        df = pd.read_csv(output_file)

        data_expected = {
            "File": ["foo/bar.txt", "spam/eggs.pdf", "testdata/bitstreams/test.pdf"],
            "Problem": ["File doesn't exist", "File doesn't exist", "Extension not supported"],
        }
        df_expected = pd.DataFrame(data_expected)
        assert df.equals(df_expected)
        output_file.unlink()



import pytest

from dsp_tools.commands.validate_data.models.api_responses import OneList


@pytest.fixture
def one_list() -> OneList:
    return OneList(list_iri="http://rdfh.ch/lists/9999/list1", list_name="firstList", nodes=["n1", "n1.1", "n1.1.1"])


class TestOneList:
    def test_hlist(self, one_list: OneList) -> None:
        result = one_list.hlist()
        expected = '"hlist=<http://rdfh.ch/lists/9999/list1>"'
        assert result == expected



from __future__ import annotations

from typing import TypeAlias
from typing import Union

from dsp_tools.xmllib import AudioSegmentResource
from dsp_tools.xmllib import LinkResource
from dsp_tools.xmllib import RegionResource
from dsp_tools.xmllib import Resource
from dsp_tools.xmllib import VideoSegmentResource

AnyResource: TypeAlias = Union[Resource, RegionResource, LinkResource, VideoSegmentResource, AudioSegmentResource]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Protocol

SCHEMA = "https://raw.githubusercontent.com/dasch-swiss/dsp-tools/main/src/dsp_tools/resources/schema/project.json"


@dataclass
class JsonHeader(Protocol):
    def to_dict(self) -> dict[str, Any]:
        raise NotImplementedError


@dataclass
class EmptyJsonHeader(JsonHeader):
    def to_dict(self) -> dict[str, Any]:
        return {
            "prefixes": {"": ""},
            "$schema": SCHEMA,
            "project": {
                "shortcode": "",
                "shortname": "",
                "longname": "",
                "descriptions": {"en": ""},
                "keywords": [""],
            },
        }


@dataclass
class FilledJsonHeader(JsonHeader):
    project: Project
    prefixes: Prefixes | None

    def to_dict(self) -> dict[str, Any]:
        header_dict: dict[str, Any] = {}
        if self.prefixes:
            header_dict["prefixes"] = self.prefixes.to_dict()
        header_dict.update({"$schema": SCHEMA, "project": self.project.to_dict()})
        return header_dict


@dataclass
class Prefixes:
    prefixes: dict[str, str]

    def to_dict(self) -> dict[str, Any]:
        return self.prefixes


@dataclass
class Project:
    shortcode: str
    shortname: str
    longname: str
    descriptions: Descriptions
    keywords: Keywords
    users: Users | None

    def to_dict(self) -> dict[str, Any]:
        proj_dict: dict[str, Any] = {
            "shortcode": self.shortcode,
            "shortname": self.shortname,
            "longname": self.longname,
            "descriptions": self.descriptions.to_dict(),
            "keywords": self.keywords.to_dict(),
        }
        if self.users:
            proj_dict["users"] = self.users.to_dict()
        return proj_dict


@dataclass
class Descriptions:
    descriptions: dict[str, str]

    def __post_init__(self) -> None:
        for k, v in self.descriptions.items():
            self.descriptions[k] = v.replace("\n", "<br/>")

    def to_dict(self) -> dict[str, str]:
        return self.descriptions


@dataclass
class Keywords:
    keywords: list[str]

    def to_dict(self) -> list[str]:
        return sorted(self.keywords)


@dataclass
class Users:
    users: list[User]

    def to_dict(self) -> list[dict[str, Any]]:
        return [x.to_dict() for x in self.users]


@dataclass
class User:
    username: str
    email: str
    givenName: str
    familyName: str
    password: str
    lang: str
    role: UserRole

    def to_dict(self) -> dict[str, Any]:
        usr_dict = {
            "username": self.username,
            "email": self.email,
            "givenName": self.givenName,
            "familyName": self.familyName,
            "password": self.password,
            "lang": self.lang,
            "status": True,
        } | self.role.to_dict()
        return usr_dict


@dataclass
class UserRole:
    project_admin: bool = False
    sys_admin: bool = False

    def to_dict(self) -> dict[str, list[str]]:
        if self.sys_admin:
            return {"groups": ["SystemAdmin"], "projects": [":admin", ":member"]}
        elif self.project_admin:
            return {"projects": [":admin", ":member"]}
        else:
            return {"projects": [":member"]}



import importlib.resources
import shutil
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import regex
import requests
import yaml
from loguru import logger

from dsp_tools.error.exceptions import InputError

MAX_FILE_SIZE = 100_000


@dataclass(frozen=True)
class StackConfiguration:
    """
    Groups together configuration information for the StackHandler.

    Args:
        max_file_size: max. multimedia file size allowed for ingest, in MB (max: 100'000)
        enforce_docker_system_prune: if True, prune Docker without asking the user
        suppress_docker_system_prune: if True, don't prune Docker (and don't ask)
        latest_dev_version: if True, start DSP-API from repo's main branch, instead of the latest deployed version
    """

    max_file_size: Optional[int] = None
    enforce_docker_system_prune: bool = False
    suppress_docker_system_prune: bool = False
    latest_dev_version: bool = False
    upload_test_data: bool = False

    def __post_init__(self) -> None:
        """
        Validate the input parameters passed by the user.

        Raises:
            InputError: if one of the parameters is invalid
        """
        if self.max_file_size is not None and not 1 <= self.max_file_size <= MAX_FILE_SIZE:
            raise InputError(f"max_file_size must be between 1 and {MAX_FILE_SIZE}")
        if self.enforce_docker_system_prune and self.suppress_docker_system_prune:
            raise InputError('The arguments "--prune" and "--no-prune" are mutually exclusive')


class StackHandler:
    """This class contains functions to start and stop the Docker containers of DSP-API and DSP-APP."""

    __stack_configuration: StackConfiguration
    __url_prefix: str
    __docker_path_of_user: Path
    __localhost_url = "http://0.0.0.0"

    def __init__(self, stack_configuration: StackConfiguration) -> None:
        """
        Initialize a StackHandler with a StackConfiguration

        Args:
            stack_configuration: configuration information for the StackHandler
        """
        self.__stack_configuration = stack_configuration
        self.__url_prefix = self._get_url_prefix()
        self.__docker_path_of_user = Path.home() / Path(".dsp-tools/start-stack")
        self.__docker_path_of_user.mkdir(parents=True, exist_ok=True)

    def _get_url_prefix(self) -> str:
        """
        The start-stack command needs some files from the DSP-API repository.
        By default, start-stack starts the latest deployed version of DSP-API.
        Since the last deployment, the DSP-API repository may have been updated.
        For this reason, we need to know the commit hash of the DSP-API version that is currently deployed,
        so that the files can be retrieved from the correct commit.

        This function reads the version tag in the docker-compose.yml file,
        and constructs the URL prefix necessary to retrieve the files from the DSP-API repository.

        If the latest development version of DSP-API is started,
        the URL prefix points to the main branch of the DSP-API repository.

        Returns:
            URL prefix used to retrieve files from the DSP-API repository
        """
        url_prefix_base = "https://raw.githubusercontent.com/dasch-swiss/dsp-api"

        if self.__stack_configuration.latest_dev_version:
            return f"{url_prefix_base}/main/"

        docker_compose_pth = importlib.resources.files("dsp_tools").joinpath("resources/start-stack/docker-compose.yml")
        docker_compose = yaml.safe_load(docker_compose_pth.read_bytes())
        tag = docker_compose["services"]["api"]["image"].split(":")[-1]

        return f"{url_prefix_base}/{tag}/"

    def _copy_resources_to_home_dir(self) -> None:
        """
        On most systems, Docker is not allowed to access files outside of the user's home directory.
        For this reason, copy the contents of the distribution (src/dsp_tools/resources/start-stack)
        to the user's home directory (~/.dsp-tools/start-stack).

        Important: The files of the home directory might have been modified
        by an earlier run of this method.
        So, this method must always be called, at every run of start-stack.
        """
        docker_path_of_distribution = importlib.resources.files("dsp_tools").joinpath("resources/start-stack")
        for file in docker_path_of_distribution.iterdir():
            with importlib.resources.as_file(file) as f:
                file_path = Path(f)
            shutil.copy(file_path, self.__docker_path_of_user / file.name)
        if not self.__stack_configuration.latest_dev_version:
            Path(self.__docker_path_of_user / "docker-compose.override.yml").unlink()

    def _get_sipi_docker_config_lua(self) -> None:
        """
        Retrieve the config file sipi.docker-config.lua from the DSP-API repository,
        and set the max_file_size parameter if necessary.

        Raises:
            InputError: if max_file_size is set but cannot be injected into sipi.docker-config.lua
        """
        docker_config_lua_response = requests.get(f"{self.__url_prefix}sipi/config/sipi.docker-config.lua", timeout=30)
        docker_config_lua_text = docker_config_lua_response.text
        if self.__stack_configuration.max_file_size:
            max_post_size_regex = r"max_post_size ?= ?[\'\"]?\d+[MG][\'\"]?"
            if not regex.search(max_post_size_regex, docker_config_lua_text):
                raise InputError("Unable to set max_file_size. Please try again without this flag.")
            docker_config_lua_text = regex.sub(
                max_post_size_regex,
                f"max_post_size = '{self.__stack_configuration.max_file_size}M'",
                docker_config_lua_text,
            )
        with open(self.__docker_path_of_user / "sipi.docker-config.lua", "w", encoding="utf-8") as f:
            f.write(docker_config_lua_text)

    def _start_up_fuseki(self) -> None:
        """
        Start up the Docker container of the fuseki database.

        Raises:
            InputError: if the database cannot be started
        """
        cmd = "docker compose up -d db".split()
        completed_process = subprocess.run(cmd, cwd=self.__docker_path_of_user, check=False)
        if not completed_process or completed_process.returncode != 0:
            msg = "Cannot start the API: Error while executing 'docker compose up -d db'"
            logger.error(f"{msg}. completed_process = '{vars(completed_process)}'")
            raise InputError(msg)

    def _wait_for_fuseki(self) -> None:
        """
        Wait up to 6 minutes, until the fuseki database is up and running.
        This function imitates the behaviour of the script dsp-api/webapi/scripts/wait-for-db.sh.
        """
        for _ in range(6 * 60):
            try:
                response = requests.get(f"{self.__localhost_url}:3030/$/server", auth=("admin", "test"), timeout=10)
                if response.ok:
                    break
            except Exception:  # noqa: BLE001 (blind-except)
                time.sleep(1)
            time.sleep(1)

    def _create_knora_test_repo(self) -> None:
        """
        Inside fuseki, create the "knora-test" repository.
        This function imitates the behaviour of the script dsp-api/webapi/scripts/fuseki-init-knora-test.sh.

        Raises:
            InputError: in case of failure
        """
        repo_template_response = requests.get(
            f"{self.__url_prefix}webapi/scripts/fuseki-repository-config.ttl.template",
            timeout=30,
        )
        repo_template = repo_template_response.text
        repo_template = repo_template.replace("@REPOSITORY@", "knora-test")
        response = requests.post(
            f"{self.__localhost_url}:3030/$/datasets",
            files={"file": ("file.ttl", repo_template, "text/turtle; charset=utf8")},
            auth=("admin", "test"),
            timeout=30,
        )
        if not response.ok:
            msg = (
                "Cannot start DSP-API: Error when creating the 'knora-test' repository. "
                "Is DSP-API perhaps running already?"
            )
            logger.error(f"{msg}. response = {vars(response)}")
            raise InputError(msg)

    def _load_data_into_repo(self) -> None:
        """
        Load some basic ontologies and data into the repository.
        This function imitates the behaviour of the script
        dsp-api/webapi/target/docker/stage/opt/docker/scripts/fuseki-init-knora-test.sh.

        Raises:
            InputError: if one of the graphs cannot be created
        """
        graph_prefix = f"{self.__localhost_url}:3030/knora-test/data?graph="
        ttl_files = [
            ("webapi/src/main/resources/knora-ontologies/knora-admin.ttl", "http://www.knora.org/ontology/knora-admin"),
            ("webapi/src/main/resources/knora-ontologies/knora-base.ttl", "http://www.knora.org/ontology/knora-base"),
            ("webapi/src/main/resources/knora-ontologies/standoff-onto.ttl", "http://www.knora.org/ontology/standoff"),
            ("webapi/src/main/resources/knora-ontologies/standoff-data.ttl", "http://www.knora.org/data/standoff"),
            ("webapi/src/main/resources/knora-ontologies/salsah-gui.ttl", "http://www.knora.org/ontology/salsah-gui"),
            ("test_data/project_data/admin-data.ttl", "http://www.knora.org/data/admin"),
            ("test_data/project_data/permissions-data.ttl", "http://www.knora.org/data/permissions"),
            ("test_data/project_ontologies/anything-onto.ttl", "http://www.knora.org/ontology/0001/anything"),
            ("test_data/project_data/anything-data.ttl", "http://www.knora.org/data/0001/anything"),
        ]
        for ttl_file, graph in ttl_files:
            ttl_response = requests.get(self.__url_prefix + ttl_file, timeout=30)
            if not ttl_response.ok:
                msg = f"Cannot start DSP-API: Error when retrieving '{self.__url_prefix + ttl_file}'"
                logger.error(f"{msg}'. response = {vars(ttl_response)}")
                raise InputError(msg)
            ttl_text = ttl_response.text
            response = requests.post(
                graph_prefix + graph,
                files={"file": ("file.ttl", ttl_text, "text/turtle; charset: utf-8")},
                auth=("admin", "test"),
                timeout=30,
            )
            if not response.ok:
                logger.error(f"Cannot start DSP-API: Error when creating graph '{graph}'. response = {vars(response)}")
                raise InputError(f"Cannot start DSP-API: Error when creating graph '{graph}'")

    def _create_admin_user(self) -> None:
        """
        This function adds the default system admin user to the database.
        The password is the hash for "test".

        Raises:
            InputError: If the user cannot be created.
        """
        graph_prefix = f"{self.__localhost_url}:3030/knora-test/data?graph="
        admin_graph = "http://www.knora.org/data/admin"
        admin_user = """
        @prefix xsd:         <http://www.w3.org/2001/XMLSchema#> .
        @prefix rdf:         <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
        @prefix knora-admin: <http://www.knora.org/ontology/knora-admin#> .
        
        <http://rdfh.ch/users/root>
        rdf:type                         knora-admin:User ;
        knora-admin:username             "root"^^xsd:string ;
        knora-admin:email                "root@example.com"^^xsd:string ;
        knora-admin:givenName            "System"^^xsd:string ;
        knora-admin:familyName           "Administrator"^^xsd:string ;
        knora-admin:password             "$2a$12$7XEBehimXN1rbhmVgQsyve08.vtDmKK7VMin4AdgCEtE4DWgfQbTK"^^xsd:string ;
        knora-admin:phone                "123456"^^xsd:string ;
        knora-admin:preferredLanguage    "en"^^xsd:string ;
        knora-admin:status               "true"^^xsd:boolean ;
        knora-admin:isInSystemAdminGroup "true"^^xsd:boolean .
        """
        response = requests.post(
            graph_prefix + admin_graph,
            files={"file": ("file.ttl", admin_user, "text/turtle; charset: utf-8")},
            auth=("admin", "test"),
            timeout=30,
        )
        if not response.ok:
            logger.error(f"Cannot start DSP-API: Error when creating the admin user. response = {vars(response)}")
            raise InputError("Cannot start DSP-API: Error when creating the admin user.")

    def _initialize_fuseki(self) -> None:
        """
        Create the "knora-test" repository and load some basic ontologies and data into it.
        """
        self._create_knora_test_repo()
        if self.__stack_configuration.upload_test_data:
            self._load_data_into_repo()
        else:
            self._create_admin_user()

    def _start_remaining_docker_containers(self) -> None:
        """
        Start the other Docker containers that are not running yet.
        (Fuseki is already running at this point.)
        """
        compose_str = "docker compose -f docker-compose.yml"
        if self.__stack_configuration.latest_dev_version:
            subprocess.run("docker compose pull".split(), cwd=self.__docker_path_of_user, check=True)
            compose_str += " -f docker-compose.override.yml"
        compose_str += " up -d"
        subprocess.run(compose_str.split(), cwd=self.__docker_path_of_user, check=True)

    def _wait_for_api(self) -> None:
        """
        Wait until the API is up and running.
        This mimicks the behaviour of the script webapi/scripts/wait-for-api.sh in the DSP-API repository.
        """
        for _ in range(6 * 60):
            try:
                response = requests.get(f"{self.__localhost_url}:3333/health", timeout=1)
                if not response.ok:
                    time.sleep(1)
                    continue
            except requests.exceptions.RequestException:
                time.sleep(1)
                continue
        print(f"DSP-API is now running on {self.__localhost_url}:3333/ and DSP-APP on {self.__localhost_url}:4200/")

    def _execute_docker_system_prune(self) -> None:
        """
        Depending on the CLI parameters or the user's input,
        execute "docker system prune" or not.
        """
        if self.__stack_configuration.enforce_docker_system_prune:
            prune_docker = "y"
        elif self.__stack_configuration.suppress_docker_system_prune:
            prune_docker = "n"
        else:
            prune_docker = None
            while prune_docker not in ["y", "n"]:
                prune_docker = input(
                    "Allow dsp-tools to execute 'docker system prune'? \n"
                    "If you press 'y', all unused containers, networks, and images (both dangling and unused) "
                    "in your docker will be deleted.\n"
                    "It is recommended that you do this every once in a while "
                    "to keep your docker clean and running smoothly. [y/n]"
                )
        if prune_docker == "y":
            subprocess.run("docker system prune --volumes -f".split(), cwd=self.__docker_path_of_user, check=False)

    def _start_docker_containers(self) -> None:
        """
        Start the fuseki Docker container,
        wait until it is up and running,
        load some basic ontologies and data into it,
        start the other Docker containers,
        and execute "docker system prune" if necessary.
        """
        self._start_up_fuseki()
        self._wait_for_fuseki()
        self._initialize_fuseki()
        self._start_remaining_docker_containers()
        self._wait_for_api()
        self._execute_docker_system_prune()

    def start_stack(self) -> bool:
        """
        Start the Docker containers of DSP-API and DSP-APP, and load some basic data models and data.
        After startup, ask user if Docker should be pruned or not.

        Raises:
            InputError: if the stack cannot be started with the parameters passed by the user

        Returns:
            True if everything went well, False otherwise
        """
        if subprocess.run("docker stats --no-stream".split(), check=False, capture_output=True).returncode != 0:
            raise InputError("Docker is not running properly. Please start Docker and try again.")
        self._copy_resources_to_home_dir()
        self._get_sipi_docker_config_lua()
        self._start_docker_containers()
        return True

    def stop_stack(self) -> bool:
        """
        Shut down the Docker containers of your local DSP stack and delete all data that is in it.

        Returns:
            True if everything went well, False otherwise
        """
        subprocess.run("docker compose down --volumes".split(), cwd=self.__docker_path_of_user, check=True)
        return True



from __future__ import annotations

from dataclasses import dataclass
from itertools import groupby

from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext


@dataclass(frozen=True)
class StandoffStashItem:
    res_id: str
    res_type: str
    value: IntermediaryRichtext


@dataclass(frozen=True)
class StandoffStash:
    """Holds information about a number of stashed XML text values, organized by resource instance."""

    res_2_stash_items: dict[str, list[StandoffStashItem]]

    @staticmethod
    def make(items: list[StandoffStashItem]) -> StandoffStash | None:
        """
        Factory method for StandoffStash.

        Args:
            items: A list of StandoffStashItem.

        Returns:
            StandoffStash | None: A StandoffStash object or None, if an empty list was passed.
        """

        def _get_res_id(x: StandoffStashItem) -> str:
            return x.res_id

        if not items:
            return None
        items = sorted(items, key=_get_res_id)
        grouped_objects = {k: list(vals) for k, vals in groupby(items, key=_get_res_id)}
        return StandoffStash(grouped_objects)


@dataclass(frozen=True)
class LinkValueStashItem:
    """Holds information about a single stashed link value."""

    res_id: str
    res_type: str
    value: IntermediaryLink


@dataclass(frozen=True)
class LinkValueStash:
    """Holds information about a number of stashed link values (resptr-props), organized by resource instance."""

    res_2_stash_items: dict[str, list[LinkValueStashItem]]

    @staticmethod
    def make(items: list[LinkValueStashItem]) -> LinkValueStash | None:
        """
        Factory method for LinkValueStash.

        Args:
            items: A list of LinkValueStashItem.

        Returns:
            LinkValueStash | None: A LinkValueStash object or None, if an empty list was passed.
        """

        def _get_res_id(x: LinkValueStashItem) -> str:
            return x.res_id

        if not items:
            return None
        items = sorted(items, key=_get_res_id)
        grouped_objects = {k: list(vals) for k, vals in groupby(items, key=_get_res_id)}
        return LinkValueStash(grouped_objects)


@dataclass(frozen=True)
class Stash:
    """Holds a standoff stash and a linkvalue stash"""

    standoff_stash: StandoffStash | None
    link_value_stash: LinkValueStash | None

    @staticmethod
    def make(standoff_stash: StandoffStash | None, link_value_stash: LinkValueStash | None) -> Stash | None:
        """
        Factory method for Stash.

        Args:
            standoff_stash: A StandoffStash object or None.
            link_value_stash: A LinkValueStash object or None.

        Returns:
            Stash: A Stash object, or None if both inputs are None.
        """
        if standoff_stash or link_value_stash:
            return Stash(standoff_stash, link_value_stash)
        return None

    def is_empty(self) -> bool:
        """Check if there are any stashed items in this stash"""
        standoff = not self.standoff_stash or not self.standoff_stash.res_2_stash_items
        link = not self.link_value_stash or not self.link_value_stash.res_2_stash_items
        return standoff and link



"""This module handles the ontology creation, update and upload to a DSP server. This includes the creation and update
of the project, the creation of groups, users, lists, resource classes, properties and cardinalities."""

from pathlib import Path
from typing import Any

from loguru import logger

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.clients.connection import Connection
from dsp_tools.clients.connection_live import ConnectionLive
from dsp_tools.commands.project.create.parse_project import parse_project_json
from dsp_tools.commands.project.create.project_create_lists import create_lists_on_server
from dsp_tools.commands.project.create.project_create_ontologies import create_ontologies
from dsp_tools.commands.project.create.project_validate import validate_project
from dsp_tools.commands.project.legacy_models.context import Context
from dsp_tools.commands.project.legacy_models.group import Group
from dsp_tools.commands.project.legacy_models.project import Project
from dsp_tools.commands.project.legacy_models.user import User
from dsp_tools.commands.project.models.project_definition import ProjectMetadata
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.legacy_models.langstring import LangString
from dsp_tools.utils.json_parsing import parse_json_input


def create_project(
    project_file_as_path_or_parsed: str | Path | dict[str, Any],
    creds: ServerCredentials,
    verbose: bool = False,
) -> bool:
    """
    Creates a project from a JSON project file on a DSP server.
    A project must contain at least one ontology,
    and it may contain lists, users, and groups.
    Severe errors lead to a BaseError,
    while other errors are printed without interrupting the process.

    Args:
        project_file_as_path_or_parsed: path to the JSON project definition, or parsed JSON object
        creds: credentials to connect to the DSP server
        verbose: prints more information if set to True

    Raises:
        InputError:
          - if the project cannot be created
          - if the login fails
          - if an ontology cannot be created

        BaseError:
          - if the input is invalid
          - if an Excel file referenced in the "lists" section cannot be expanded
          - if the validation doesn't pass

    Returns:
        True if everything went smoothly, False if a warning or error occurred
    """

    knora_api_prefix = "knora-api:"
    overall_success = True

    project_json = parse_json_input(project_file_as_path_or_parsed=project_file_as_path_or_parsed)

    context = Context(project_json.get("prefixes", {}))

    # validate against JSON schema
    validate_project(project_json)
    print("    JSON project file is syntactically correct and passed validation.")
    logger.info("JSON project file is syntactically correct and passed validation.")

    project = parse_project_json(project_json)

    auth = AuthenticationClientLive(creds.server, creds.user, creds.password)
    con = ConnectionLive(creds.server, auth)

    # create project on DSP server
    info_str = f"Create project '{project.metadata.shortname}' ({project.metadata.shortcode})..."
    print(info_str)
    logger.info(info_str)
    project_remote, success = _create_project_on_server(
        project_definition=project.metadata,
        con=con,
    )
    if not success:
        overall_success = False

    # create the lists
    names_and_iris_of_list_nodes: dict[str, Any] = {}
    if project.lists:
        print("Create lists...")
        logger.info("Create lists...")
        names_and_iris_of_list_nodes, success = create_lists_on_server(
            lists_to_create=project.lists,
            con=con,
            project_remote=project_remote,
        )
        if not success:
            overall_success = False

    # create the groups
    current_project_groups: dict[str, Group] = {}
    if project.groups:
        print("Create groups...")
        logger.info("Create groups...")
        current_project_groups, success = _create_groups(
            con=con,
            groups=project.groups,
            project=project_remote,
        )
        if not success:
            overall_success = False

    # create or update the users
    if project.users:
        print("Create users...")
        logger.info("Create users...")
        success = _create_users(
            con=con,
            users_section=project.users,
            current_project_groups=current_project_groups,
            current_project=project_remote,
            verbose=verbose,
        )
        if not success:
            overall_success = False

    # create the ontologies
    success = create_ontologies(
        con=con,
        context=context,
        knora_api_prefix=knora_api_prefix,
        names_and_iris_of_list_nodes=names_and_iris_of_list_nodes,
        ontology_definitions=project.ontologies,
        project_remote=project_remote,
        verbose=verbose,
    )
    if not success:
        overall_success = False

    # final steps
    if overall_success:
        msg = (
            f"Successfully created project '{project.metadata.shortname}' "
            f"({project.metadata.shortcode}) with all its ontologies. "
            f"There were no problems during the creation process."
        )
        print(f"========================================================\n{msg}")
        logger.info(msg)
    else:
        msg = (
            f"The project '{project.metadata.shortname}' ({project.metadata.shortcode}) "
            f"with its ontologies could be created, "
            f"but during the creation process, some problems occurred. Please carefully check the console output."
        )
        print(f"========================================================\nWARNING: {msg}")
        logger.warning(msg)

    return overall_success


def _create_project_on_server(
    project_definition: ProjectMetadata,
    con: Connection,
) -> tuple[Project, bool]:
    """
    Create the project on the DSP server.
    If it already exists: update its longname, description, and keywords.

    Args:
        project_definition: object with information about the project
        con: connection to the DSP server

    Raises:
        InputError: if the project cannot be created on the DSP server

    Returns:
        a tuple of the remote project and the success status (True if everything went smoothly, False otherwise)
    """
    all_projects = Project.getAllProjects(con=con)
    if project_definition.shortcode in [proj.shortcode for proj in all_projects]:
        msg = (
            f"The project with the shortcode '{project_definition.shortcode}' already exists on the server.\n"
            f"No changes were made to the project metadata.\n"
            f"Continue with the upload of lists and ontologies ..."
        )
        print(f"WARNING: {msg}")
        logger.warning(msg)

    success = True
    project_local = Project(
        con=con,
        shortcode=project_definition.shortcode,
        shortname=project_definition.shortname,
        longname=project_definition.longname,
        description=LangString(project_definition.descriptions),  # type: ignore[arg-type]
        keywords=set(project_definition.keywords) if project_definition.keywords else None,
        selfjoin=False,
        status=True,
    )
    try:
        project_remote = project_local.create()
    except BaseError:
        err_msg = (
            f"Cannot create project '{project_definition.shortname}' ({project_definition.shortcode}) on DSP server."
        )
        logger.exception(err_msg)
        raise InputError(err_msg) from None
    print(f"    Created project '{project_remote.shortname}' ({project_remote.shortcode}).")
    logger.info(f"Created project '{project_remote.shortname}' ({project_remote.shortcode}).")
    return project_remote, success


def _create_groups(
    con: Connection,
    groups: list[dict[str, str]],
    project: Project,
) -> tuple[dict[str, Group], bool]:
    """
    Creates groups on a DSP server from the "groups" section of a JSON project file. If a group cannot be created, it is
    skipped and a warning is printed, but such a group will still be part of the returned dict.
    Returns a tuple consisting of a dict and a bool. The dict contains the groups that have successfully been created
    (or already exist). The bool indicates if everything went smoothly during the process. If a warning or error
    occurred, it is False.

    Args:
        con: connection instance to connect to the DSP server
        groups: "groups" section of a parsed JSON project file
        project: Project the group(s) should be added to (must exist on DSP server)

    Returns:
        A tuple consisting of a dict and the success status.
        The dict has the form ``{group name: group object}``
        for all groups that have successfully been created (or already exist).
        The dict is empty if no group was created.
    """
    overall_success = True
    current_project_groups: dict[str, Group] = {}
    try:
        remote_groups = Group.getAllGroupsForProject(con=con, proj_iri=str(project.iri))
    except BaseError:
        err_msg = (
            "Unable to check if group names are already existing on DSP server, because it is "
            "not possible to retrieve the remote groups from the DSP server."
        )
        print(f"WARNING: {err_msg}")
        logger.exception(err_msg)
        remote_groups = []
        overall_success = False

    for group in groups:
        group_name = group["name"]

        # if the group already exists, add it to "current_project_groups" (for later usage), then skip it
        if remotely_existing_group := [g for g in remote_groups if g.name == group_name]:
            current_project_groups[group_name] = remotely_existing_group[0]
            err_msg = f"Group name '{group_name}' already exists on the DSP server. Skipping..."
            print(f"    WARNING: {err_msg}")
            logger.warning(err_msg)
            overall_success = False
            continue

        # create the group
        group_local = Group(
            con=con,
            name=group_name,
            descriptions=LangString(group["descriptions"]),
            project=project,
            status=bool(group.get("status", True)),
            selfjoin=bool(group.get("selfjoin", False)),
        )
        try:
            group_remote: Group = group_local.create()
        except BaseError:
            err_msg = "Unable to create group '{group_name}'."
            print(f"    WARNING: {err_msg}")
            logger.exception(err_msg)
            overall_success = False
            continue

        current_project_groups[str(group_remote.name)] = group_remote
        print(f"    Created group '{group_name}'.")
        logger.info(f"Created group '{group_name}'.")

    return current_project_groups, overall_success


def _create_users(
    con: Connection,
    users_section: list[dict[str, str]],
    current_project_groups: dict[str, Group],
    current_project: Project,
    verbose: bool,
) -> bool:
    """
    Creates users on a DSP server from the "users" section of a JSON project file.
    If a user cannot be created, a warning is printed and the user is skipped.

    Args:
        con: connection instance to connect to the DSP server
        users_section: "users" section of a parsed JSON project file
        current_project_groups: groups defined in the current project, in the form ``{group name: group object}``
            (must exist on DSP server)
        current_project: "project" object of the current project (must exist on DSP server)
        verbose: Prints more information if set to True

    Returns:
        True if all users could be created without any problems. False if a warning/error occurred.
    """
    overall_success = True
    for json_user_definition in users_section:
        username = json_user_definition["username"]

        # skip the user if he already exists
        all_users = User.getAllUsers(con)
        if json_user_definition["email"] in [user.email for user in all_users]:
            err_msg = (
                f"User '{username}' already exists on the DSP server.\n"
                f"Please manually add this user to the project in DSP-APP."
            )
            print(f"    WARNING: {err_msg}")
            logger.warning(err_msg)
            overall_success = False
            continue
        # add user to the group(s)
        group_iris, sysadmin, success = _get_group_iris_for_user(
            json_user_definition=json_user_definition,
            current_project=current_project,
            current_project_groups=current_project_groups,
            con=con,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # add user to the project(s)
        project_info, success = _get_projects_where_user_is_admin(
            json_user_definition=json_user_definition,
            current_project=current_project,
            con=con,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # create the user
        user_local = User(
            con=con,
            username=json_user_definition["username"],
            email=json_user_definition["email"],
            givenName=json_user_definition["givenName"],
            familyName=json_user_definition["familyName"],
            password=json_user_definition["password"],
            status=bool(json_user_definition.get("status", True)),
            lang=json_user_definition.get("lang", "en"),
            sysadmin=sysadmin,
            in_projects=project_info,
            in_groups=group_iris,
        )
        try:
            user_local.create()
        except BaseError:
            print(f"    WARNING: Unable to create user '{username}'.")
            logger.exception(f"Unable to create user '{username}'.")
            overall_success = False
            continue
        print(f"    Created user '{username}'.")
        logger.info(f"Created user '{username}'.")

    return overall_success


def _get_group_iris_for_user(
    json_user_definition: dict[str, str],
    current_project: Project,
    current_project_groups: dict[str, Group],
    con: Connection,
    verbose: bool,
) -> tuple[set[str], bool, bool]:
    """
    Retrieve the IRIs of the groups that the user belongs to.

    Args:
        json_user_definition: the section of the JSON file that defines a user
        current_project: the Project object
        current_project_groups: dict of the form ``{group name: group object}``
            with the groups that exist on the DSP server
        con: connection to the DSP server
        verbose: verbose switch

    Returns:
        a tuple consisting of the group IRIs,
        the system admin status (True if the user is sysadmin, False otherwise),
        and the success status (True if everything went well)

    Raises:
        BaseError: if no groups can be retrieved from the DSP server, or if the retrieved group has no IRI
    """
    success = True
    username = json_user_definition["username"]
    group_iris: set[str] = set()
    sysadmin = False
    remote_groups: list[Group] = []
    for full_group_name in json_user_definition.get("groups", []):
        # full_group_name has the form '[project_shortname]:group_name' or 'SystemAdmin'
        inexisting_group_msg = (
            f"User {username} cannot be added to group {full_group_name}, because such a group doesn't exist."
        )
        if ":" not in full_group_name and full_group_name != "SystemAdmin":
            print(f"    WARNING: {inexisting_group_msg}")
            logger.warning(inexisting_group_msg)
            success = False
            continue

        if full_group_name == "SystemAdmin":
            sysadmin = True
            if verbose:
                print(f"    Added user '{username}' to group 'SystemAdmin'.")
            logger.info(f"Added user '{username}' to group 'SystemAdmin'.")
            continue

        # all other cases (":" in full_group_name)
        project_shortname, group_name = full_group_name.split(":")
        if not project_shortname:
            # full_group_name refers to a group inside the same project
            if group_name not in current_project_groups:
                print(f"    WARNING: {inexisting_group_msg}")
                logger.warning(inexisting_group_msg)
                success = False
                continue
            group = current_project_groups[group_name]
        else:
            # full_group_name refers to an already existing group on DSP
            try:
                # "remote_groups" might be available from a previous loop cycle
                remote_groups = remote_groups or Group.getAllGroups(con=con)
            except BaseError:
                err_msg = (
                    f"User '{username}' is referring to the group {full_group_name} that "
                    f"exists on the DSP server, but no groups could be retrieved from the DSP server."
                )
                print(f"    WARNING: {err_msg}")
                logger.exception(err_msg)
                success = False
                continue
            existing_group = [g for g in remote_groups if g.project == current_project.iri and g.name == group_name]
            if not existing_group:
                print(f"    WARNING: {inexisting_group_msg}")
                logger.warning(inexisting_group_msg)
                success = False
                continue
            group = existing_group[0]

        if not group.iri:
            raise BaseError(f"Group '{group}' has no IRI.")
        group_iris.add(group.iri)
        if verbose:
            print(f"    Added user '{username}' to group '{full_group_name}'.")
        logger.info(f"Added user '{username}' to group '{full_group_name}'.")

    return group_iris, sysadmin, success


def _get_projects_where_user_is_admin(
    json_user_definition: dict[str, str],
    current_project: Project,
    con: Connection,
    verbose: bool,
) -> tuple[dict[str, bool], bool]:
    """
    Create a dict that tells for every project if the user is administrator in that project or not.

    Args:
        json_user_definition: the section of the JSON file that defines a user
        current_project: the Project object
        con: connection to the DSP server
        verbose: verbose switch

    Returns:
        a tuple consisting of a dict in the form {project IRI: isAdmin}, and the success status
    """
    success = True
    username = json_user_definition["username"]
    project_info: dict[str, bool] = {}
    remote_projects: list[Project] = []
    for full_project_name in json_user_definition.get("projects", []):
        # full_project_name has the form '[project_name]:member' or '[project_name]:admin'
        if ":" not in full_project_name:
            err_msg = "Provided project '{full_project_name}' for user '{username}' is not valid. Skipping..."
            print(f"    WARNING: {err_msg}")
            logger.warning(err_msg)
            success = False
            continue

        project_name, project_role = full_project_name.split(":")
        if not project_name:
            # full_project_name refers to the current project
            in_project = current_project
        else:
            # full_project_name refers to an already existing project on DSP
            try:
                # "remote_projects" might be available from a previous loop cycle
                remote_projects = remote_projects or current_project.getAllProjects(con=con)
            except BaseError:
                err_msg = (
                    f"User '{username}' cannot be added to the projects {json_user_definition['projects']} "
                    f"because the projects cannot be retrieved from the DSP server."
                )
                print(f"    WARNING: {err_msg}")
                logger.exception(err_msg)
                success = False
                continue
            in_project_list = [p for p in remote_projects if p.shortname == project_name]
            if not in_project_list:
                msg = f"Provided project '{full_project_name}' for user '{username}' is not valid. Skipping..."
                print(f"    WARNING: {msg}")
                logger.warning(msg)
                success = False
                continue
            in_project = in_project_list[0]

        is_admin = project_role == "admin"
        match project_info.get(str(in_project.iri)):
            case None:  # user hasn't yet been registered as admin/member of this project
                project_info[str(in_project.iri)] = is_admin
            case True:  # user has already been registered as **admin** of this project
                pass  # leave previous admin/member status untouched
            case False:  # user has already been registered as **member** of this project
                project_info[str(in_project.iri)] = is_admin  # overwrite previous admin/member status
        if verbose:
            print(f"    Added user '{username}' as {project_role} to project '{in_project.shortname}'.")
        logger.info(f"Added user '{username}' as {project_role} to project '{in_project.shortname}'.")

    return project_info, success



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any


@dataclass
class ListNode:
    id_: str
    labels: dict[str, str]
    comments: dict[str, str]
    parent_id: str
    sub_nodes: list[ListNode] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        node = self._make_own_node()
        if self.sub_nodes:
            node["nodes"] = self._make_subnodes()
        return node

    def _make_subnodes(self) -> list[dict[str, Any]]:
        return [nd.to_dict() for nd in self.sub_nodes]

    def _make_own_node(self) -> dict[str, Any]:
        d = {"name": self.id_, "labels": self.labels}
        if self.comments:
            d.update({"comments": self.comments})
        return d


@dataclass
class ListRoot:
    id_: str
    labels: dict[str, str]
    nodes: list[ListNode]
    comments: dict[str, str]

    def to_dict(self) -> dict[str, Any]:
        lst_root = self._make_list_root()
        lst_root["nodes"] = self._make_nodes()
        return lst_root

    def _make_nodes(self) -> list[dict[str, Any]]:
        return [nd.to_dict() for nd in self.nodes]

    def _make_list_root(self) -> dict[str, Any]:
        lst = {"name": self.id_, "labels": self.labels}
        if self.comments:
            lst.update({"comments": self.comments})
        else:
            lst.update({"comments": self.labels})
        return lst



from typing import Protocol

# ruff: noqa: D102 (missing docstring in public method)


class AuthenticationClient(Protocol):
    """
    Protocol for a client that can authenticate with a DSP server and returns a token.
    """

    def get_token(self) -> str:
        pass



import pytest

from dsp_tools.commands.excel2json.lists.models.serialise import ListNode
from dsp_tools.commands.excel2json.lists.models.serialise import ListRoot


class TestListRoot:
    def test_make_myself_comments(self) -> None:
        sub_1 = ListNode(id_="Node_1", labels={"en": "Node_1_label_en"}, comments={}, parent_id="RootID")
        root = ListRoot(
            id_="RootID",
            labels={"en": "root_label_en", "de": "root_label_de"},
            nodes=[sub_1],
            comments={"en": "root_comment_en", "de": "root_comment_de"},
        )
        expected = {
            "name": "RootID",
            "labels": {"en": "root_label_en", "de": "root_label_de"},
            "comments": {"en": "root_comment_en", "de": "root_comment_de"},
        }
        res = root._make_list_root()
        assert expected == res

    def test_make_myself_no_comments(self) -> None:
        sub_1 = ListNode(id_="Node_1", labels={"en": "Node_1_label_en"}, comments={}, parent_id="RootID")
        root = ListRoot(
            id_="RootID",
            labels={"en": "root_label_en", "de": "root_label_de"},
            comments={"en": "root_comment_en", "de": "root_comment_de"},
            nodes=[sub_1],
        )
        expected = {
            "name": "RootID",
            "labels": {"en": "root_label_en", "de": "root_label_de"},
            "comments": {"en": "root_comment_en", "de": "root_comment_de"},
        }
        res = root._make_list_root()
        assert expected == res

    def test_to_dict(self) -> None:
        sub_1 = ListNode(id_="Node_1", labels={"en": "Node_1_label_en"}, comments={}, parent_id="RootID")
        sub_21 = ListNode(
            id_="SubNode_21", labels={"en": "SubNode_21_label_en"}, comments={"en": "en_comment"}, parent_id="RootID"
        )
        sub_2 = ListNode(
            id_="Node_2", labels={"en": "Node_2_label_en"}, comments={}, sub_nodes=[sub_21], parent_id="RootID"
        )
        root = ListRoot(
            id_="RootID",
            labels={"en": "root_label_en", "de": "root_label_de"},
            comments={},
            nodes=[sub_1, sub_2],
        )
        expected = {
            "name": "RootID",
            "labels": {"en": "root_label_en", "de": "root_label_de"},
            "comments": {"en": "root_label_en", "de": "root_label_de"},
            "nodes": [
                {"name": "Node_1", "labels": {"en": "Node_1_label_en"}},
                {
                    "name": "Node_2",
                    "labels": {"en": "Node_2_label_en"},
                    "nodes": [
                        {
                            "name": "SubNode_21",
                            "labels": {"en": "SubNode_21_label_en"},
                            "comments": {"en": "en_comment"},
                        }
                    ],
                },
            ],
        }
        res = root.to_dict()
        assert res == expected


class TestListNode:
    def test_make_myself(self) -> None:
        nd = ListNode(id_="NodeID", labels={"en": "node_label_en"}, comments={}, parent_id="RootID")
        expected = {"name": "NodeID", "labels": {"en": "node_label_en"}}
        res = nd._make_own_node()
        assert res == expected

    def test_to_dict(self) -> None:
        sub_1 = ListNode(
            id_="SubNode_1", labels={"en": "SubNode_1_label_en"}, comments={"en": "comment en"}, parent_id="RootID"
        )
        sub_21 = ListNode(id_="SubNode_21", labels={"en": "SubNode_21_label_en"}, comments={}, parent_id="RootID")
        sub_2 = ListNode(
            id_="SubNode_2", labels={"en": "SubNode_2_label_en"}, comments={}, sub_nodes=[sub_21], parent_id="RootID"
        )
        test_nd = ListNode(
            id_="NodeID", labels={"en": "node_label_en"}, comments={}, sub_nodes=[sub_1, sub_2], parent_id="RootID"
        )

        expected = {
            "name": "NodeID",
            "labels": {"en": "node_label_en"},
            "nodes": [
                {"name": "SubNode_1", "labels": {"en": "SubNode_1_label_en"}, "comments": {"en": "comment en"}},
                {
                    "name": "SubNode_2",
                    "labels": {"en": "SubNode_2_label_en"},
                    "nodes": [
                        {
                            "name": "SubNode_21",
                            "labels": {"en": "SubNode_21_label_en"},
                        }
                    ],
                },
            ],
        }
        res = test_nd.to_dict()
        assert res == expected


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.models.data_deserialised import MigrationMetadata
from dsp_tools.utils.xml_parsing.models.data_deserialised import PropertyObject
from dsp_tools.utils.xml_parsing.models.data_deserialised import ResourceDeserialised
from dsp_tools.utils.xml_parsing.models.data_deserialised import TripleObjectType
from dsp_tools.utils.xml_parsing.models.data_deserialised import TriplePropertyType
from dsp_tools.utils.xml_parsing.models.data_deserialised import ValueInformation
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType

LABEL_TRIPLE = PropertyObject(TriplePropertyType.RDFS_LABEL, "lbl", TripleObjectType.STRING)
TYPE_TRIPLE = PropertyObject(
    TriplePropertyType.RDF_TYPE, "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything", TripleObjectType.IRI
)

UNREIFIED_TRIPLE_OBJECTS = [LABEL_TRIPLE, TYPE_TRIPLE]


@pytest.fixture
def resource_deserialised_with_values(
    boolean_value_deserialised_corr: ValueInformation,
) -> ResourceDeserialised:
    return ResourceDeserialised(
        res_id="id",
        property_objects=UNREIFIED_TRIPLE_OBJECTS,
        values=[boolean_value_deserialised_corr],
        asset_value=None,
        migration_metadata=MigrationMetadata(),
    )


@pytest.fixture
def resource_deserialised_no_values() -> ResourceDeserialised:
    return ResourceDeserialised(
        res_id="id",
        property_objects=UNREIFIED_TRIPLE_OBJECTS,
        values=[],
        asset_value=None,
        migration_metadata=MigrationMetadata(),
    )


@pytest.fixture
def resource_deserialised_with_asset() -> ResourceDeserialised:
    return ResourceDeserialised(
        res_id="id",
        property_objects=UNREIFIED_TRIPLE_OBJECTS,
        values=[],
        asset_value=ValueInformation(
            f"{KNORA_API_STR}hasAudioFileValue",
            "testdata/bitstreams/test.wav",
            KnoraValueType.AUDIO_FILE,
            [],
        ),
        migration_metadata=MigrationMetadata(),
    )


@pytest.fixture
def boolean_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", "false", KnoraValueType.BOOLEAN_VALUE, []
    )


@pytest.fixture
def boolean_value_deserialised_zero() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", "0", KnoraValueType.BOOLEAN_VALUE, []
    )


@pytest.fixture
def boolean_value_deserialised_one() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", "1", KnoraValueType.BOOLEAN_VALUE, []
    )


@pytest.fixture
def color_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testColor", "#00ff00", KnoraValueType.COLOR_VALUE, []
    )


@pytest.fixture
def date_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1",
        "JULIAN:BCE:0700:BCE:0600",
        KnoraValueType.DATE_VALUE,
        [],
    )


@pytest.fixture
def decimal_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText",
        "1.2",
        KnoraValueType.DECIMAL_VALUE,
        [],
    )


@pytest.fixture
def geoname_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname",
        "1241345",
        KnoraValueType.GEONAME_VALUE,
        [],
    )


@pytest.fixture
def int_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText",
        "1",
        KnoraValueType.INT_VALUE,
        [],
    )


@pytest.fixture
def interval_value_deserialised_corr() -> ValueInformation:
    seg_start = PropertyObject(TriplePropertyType.KNORA_INTERVAL_START, "1", TripleObjectType.DECIMAL)
    seg_end = PropertyObject(TriplePropertyType.KNORA_INTERVAL_END, "2", TripleObjectType.DECIMAL)
    return ValueInformation(
        user_facing_prop=f"{KNORA_API_STR}hasSegmentBounds",
        user_facing_value=None,
        knora_type=KnoraValueType.INTERVAL_VALUE,
        value_metadata=[seg_start, seg_end],
    )


@pytest.fixture
def link_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo",
        "link-id",
        KnoraValueType.LINK_VALUE,
        [],
    )


@pytest.fixture
def link_value_deserialised_none() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo", None, KnoraValueType.LINK_VALUE, []
    )


@pytest.fixture
def list_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp", "n1", KnoraValueType.LIST_VALUE, []
    )


@pytest.fixture
def simple_text_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea",
        "simple text",
        KnoraValueType.SIMPLETEXT_VALUE,
        [],
    )


@pytest.fixture
def richtext_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext",
        "rich text",
        KnoraValueType.RICHTEXT_VALUE,
        [],
    )


@pytest.fixture
def time_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue",
        "2019-10-23T13:45:12.01-14:00",
        KnoraValueType.TIME_VALUE,
        [],
    )


@pytest.fixture
def uri_value_deserialised_corr() -> ValueInformation:
    return ValueInformation(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue",
        "https://dasch.swiss",
        KnoraValueType.URI_VALUE,
        [],
    )



from copy import deepcopy
from uuid import uuid4

import pytest

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediarySimpleText
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.commands.xmlupload.stash.stash_circular_references import stash_circular_references
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStash
from dsp_tools.commands.xmlupload.stash.stash_models import Stash

PERMISSION = Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})


@pytest.fixture
def resource_1() -> IntermediaryResource:
    link_value_to_resource_no_links = IntermediaryLink("res_no_links", "prop", None, None, str(uuid4()))
    link_value_with_permissions_to_res_2 = IntermediaryLink("res_2", "prop", None, PERMISSION, "link_to_res_2_uuid")
    simple_text_value = IntermediarySimpleText("Text", "prop", None, None)
    return IntermediaryResource(
        res_id="res_1",
        type_iri="type",
        label="lbl",
        permissions=None,
        values=[link_value_with_permissions_to_res_2, link_value_to_resource_no_links, simple_text_value],
    )


@pytest.fixture
def resource_2() -> IntermediaryResource:
    val = IntermediaryRichtext(
        FormattedTextValue('Link: <a class="salsah-link" href="IRI:res_3:IRI">res_3</a>'),
        "prop",
        None,
        None,
        set("res_3"),
        "standoff_link_to_res_3_uuid",
    )
    return IntermediaryResource(
        res_id="res_2",
        type_iri="type",
        label="lbl",
        permissions=None,
        values=[val],
    )


@pytest.fixture
def resource_3() -> IntermediaryResource:
    link_value_to_res_1 = IntermediaryLink("res_1", "prop", None, None, "link_to_res_1_uuid")
    return IntermediaryResource(
        res_id="res_3",
        type_iri="type",
        label="lbl",
        permissions=None,
        values=[link_value_to_res_1],
    )


@pytest.fixture
def resource_no_links() -> IntermediaryResource:
    return IntermediaryResource(
        res_id="res_no_links",
        type_iri="type",
        label="lbl",
        permissions=None,
        values=[],
    )


def test_stash_circular_references_remove_link_value(
    resource_1: IntermediaryResource,
    resource_2: IntermediaryResource,
    resource_3: IntermediaryResource,
    resource_no_links: IntermediaryResource,
) -> None:
    copied_res_1 = deepcopy(resource_1)
    copied_res_2 = deepcopy(resource_2)
    copied_res_3 = deepcopy(resource_3)
    resources = [copied_res_1, copied_res_2, copied_res_3, deepcopy(resource_no_links)]
    lookup = {"res_1": ["link_to_res_2_uuid"]}
    result = stash_circular_references(resources, lookup)
    assert isinstance(result, Stash)
    assert not result.standoff_stash
    link_stash = result.link_value_stash
    assert isinstance(link_stash, LinkValueStash)
    assert link_stash.res_2_stash_items.keys() == lookup.keys()
    stash_list = link_stash.res_2_stash_items["res_1"]
    assert len(stash_list) == 1
    stash_item = stash_list.pop(0)
    assert stash_item.res_id == "res_1"
    assert stash_item.res_type == "type"
    assert stash_item.value.prop_iri == "propValue"
    assert stash_item.value.value == "res_2"
    assert str(stash_item.value.permissions) == str(PERMISSION)

    # check that the resource values are as expected
    assert len(copied_res_1.values) == 2
    content = {x.value for x in copied_res_1.values}
    assert content == {"res_no_links", "Text"}

    assert len(resource_2.values) == len(copied_res_2.values)
    assert len(resource_3.values) == len(copied_res_3.values)


def test_stash_circular_references_remove_text_value(
    resource_1: IntermediaryResource,
    resource_2: IntermediaryResource,
    resource_3: IntermediaryResource,
    resource_no_links: IntermediaryResource,
) -> None:
    copied_res_1 = deepcopy(resource_1)
    copied_res_2 = deepcopy(resource_2)
    copied_res_3 = deepcopy(resource_3)
    resources = [copied_res_1, copied_res_2, copied_res_3, deepcopy(resource_no_links)]
    lookup = {"res_2": ["standoff_link_to_res_3_uuid"]}
    result = stash_circular_references(resources, lookup)
    assert isinstance(result, Stash)
    assert not result.link_value_stash
    standoff_stash = result.standoff_stash
    assert isinstance(standoff_stash, StandoffStash)
    assert standoff_stash.res_2_stash_items.keys() == lookup.keys()
    stash_list = standoff_stash.res_2_stash_items["res_2"]
    assert len(stash_list) == 1
    stash_item = stash_list.pop(0)
    assert stash_item.res_id == "res_2"
    assert stash_item.res_type == "type"
    assert stash_item.value.value_uuid == "standoff_link_to_res_3_uuid"
    assert stash_item.value.prop_iri == "prop"
    assert stash_item.value.value.xmlstr == 'Link: <a class="salsah-link" href="IRI:res_3:IRI">res_3</a>'

    # check that the resource values are as expected
    assert len(copied_res_2.values) == 1
    text_val = copied_res_2.values.pop(0)
    assert isinstance(text_val, IntermediaryRichtext)
    assert text_val.value.xmlstr == "standoff_link_to_res_3_uuid"
    assert len(resource_1.values) == len(copied_res_1.values)
    assert len(resource_3.values) == len(copied_res_3.values)


def test_stash_circular_references_no_stash(
    resource_1: IntermediaryResource,
) -> None:
    resources = [deepcopy(resource_1)]
    lookup: dict[str, list[str]] = {}
    assert not stash_circular_references(resources, lookup)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
import regex
from lxml import etree

from dsp_tools.error.custom_warnings import DspToolsUserInfo
from dsp_tools.utils.xml_parsing.xml_schema_validation import _warn_user_about_tags_in_simpletext


def test_warn_user_about_tags_in_simpletext_all_good(recwarn: pytest.WarningsRecorder) -> None:
    allowed_html_escapes = [
        "(&lt;2cm) (&gt;10cm)",
        "text &lt; text/&gt;",
        "text &lt; text&gt; &amp; text",
        "text &lt;text text &gt; text",
        'text &lt; text text="text"&gt; text',
        'text &lt;text text="text" &gt; text',
    ]
    utf8_texts_with_allowed_html_escapes = [
        f"""
        <knora shortcode="4123" default-ontology="testonto">
            <resource label="label" restype=":restype" id="id">
                <text-prop name=":name">
                    <text encoding="utf8">{txt}</text>
                </text-prop>
            </resource>
        </knora>
        """
        for txt in allowed_html_escapes
    ]
    for xml in utf8_texts_with_allowed_html_escapes:
        _warn_user_about_tags_in_simpletext(etree.fromstring(xml))
    assert len(recwarn) == 0


def test_warn_user_about_tags_in_simpletext_forbidden_escapes() -> None:
    test_ele = etree.fromstring(
        """
        <knora shortcode="4123" default-ontology="testonto">
            <resource label="label" restype=":restype" id="id">
                <text-prop name=":name">
                    <text encoding="utf8">&lt;tag s="t"&gt;</text>
                </text-prop>
            </resource>
        </knora>
        """
    )
    expected_msg = regex.escape(
        "Angular brackets in the format of <text> were found in text properties with encoding=utf8.\n"
        "Please note that these will not be recognised as formatting in the text field, "
        "but will be displayed as-is.\n"
        "The following resources of your XML file contain angular brackets:\n"
        "    - line 5: resource 'id', property ':name'"
    )
    with pytest.warns(DspToolsUserInfo, match=expected_msg):
        _warn_user_about_tags_in_simpletext(test_ele)


def test_warn_user_about_tags_in_simpletext_forbidden_escapes_two() -> None:
    test_ele = etree.fromstring(
        """
        <knora shortcode="4123" default-ontology="testonto">
            <resource label="label" restype=":restype" id="id">
                <text-prop name=":propName">
                    <text encoding="utf8">&lt;em&gt;text&lt;/em&gt;</text>
                </text-prop>
            </resource>
        </knora>
        """
    )
    expected_msg = regex.escape(
        "Angular brackets in the format of <text> were found in text properties with encoding=utf8.\n"
        "Please note that these will not be recognised as formatting in the text field, "
        "but will be displayed as-is.\n"
        "The following resources of your XML file contain angular brackets:\n"
        "    - line 5: resource 'id', property ':propName'"
    )

    with pytest.warns(DspToolsUserInfo, match=expected_msg):
        _warn_user_about_tags_in_simpletext(test_ele)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from datetime import datetime
from typing import cast

import regex
from loguru import logger
from lxml import etree

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.xmlupload.models.deserialise.xmlpermission import XmlPermission
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.lookup_models import IntermediaryLookups
from dsp_tools.commands.xmlupload.models.lookup_models import make_namespace_dict_from_onto_names
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.upload_clients import UploadClients
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_xmlresource_into_intermediary_classes import (
    transform_xmlresources_into_intermediary_resources,
)
from dsp_tools.commands.xmlupload.stash.analyse_circular_reference_graph import generate_upload_order
from dsp_tools.commands.xmlupload.stash.create_info_for_graph import create_info_for_graph_from_intermediary_resources
from dsp_tools.commands.xmlupload.stash.stash_circular_references import stash_circular_references
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.legacy_models.projectContext import ProjectContext

LIST_SEPARATOR = "\n-    "


def get_intermediary_lookups(root: etree._Element, con: Connection, clients: UploadClients) -> IntermediaryLookups:
    proj_context = _get_project_context_from_server(connection=con, shortcode=root.attrib["shortcode"])
    permissions_lookup = _get_permissions_lookup(root, proj_context)
    authorship_lookup = _get_authorship_lookup(root)
    listnode_lookup = clients.list_client.get_list_node_id_to_iri_lookup()
    project_onto_dict = clients.project_client.get_ontology_name_dict()
    namespaces = make_namespace_dict_from_onto_names(project_onto_dict)
    return IntermediaryLookups(
        permissions=permissions_lookup,
        listnodes=listnode_lookup,
        namespaces=namespaces,
        authorships=authorship_lookup,
    )


def _get_permissions_lookup(root: etree._Element, proj_context: ProjectContext) -> dict[str, Permissions]:
    permission_ele = list(root.iter(tag="permissions"))
    permissions = [XmlPermission(permission, proj_context) for permission in permission_ele]
    permissions_dict = {permission.permission_id: permission for permission in permissions}
    permissions_lookup = {name: perm.get_permission_instance() for name, perm in permissions_dict.items()}
    return permissions_lookup


def _get_project_context_from_server(connection: Connection, shortcode: str) -> ProjectContext:
    try:
        proj_context = ProjectContext(con=connection, shortcode=shortcode)
    except BaseError:
        logger.exception("Unable to retrieve project context from DSP server")
        raise InputError("Unable to retrieve project context from DSP server") from None
    return proj_context


def _get_authorship_lookup(root: etree._Element) -> dict[str, list[str]]:
    def get_one_author(ele: etree._Element) -> str:
        # The xsd file ensures that the body of the element contains valid non-whitespace characters
        txt = cast(str, ele.text)
        txt = regex.sub(r"[\n\t]", " ", txt)
        txt = regex.sub(r" +", " ", txt)
        return txt.strip()

    authorship_lookup = {}
    for auth in root.iter(tag="authorship"):
        individual_authors = [get_one_author(child) for child in auth.iterchildren()]
        authorship_lookup[auth.attrib["id"]] = individual_authors
    return authorship_lookup


def prepare_upload_from_root(
    root: etree._Element, default_ontology: str, intermediary_lookups: IntermediaryLookups
) -> tuple[list[IntermediaryResource], Stash | None]:
    """Do the consistency check, resolve circular references, and return the resources and permissions."""
    logger.info("Get data from XML...")
    resources = _extract_resources_from_xml(root, default_ontology)
    transformed_resources = _get_transformed_resources(resources, intermediary_lookups)
    info_for_graph = create_info_for_graph_from_intermediary_resources(transformed_resources)
    stash_lookup, upload_order = generate_upload_order(info_for_graph)
    sorting_lookup = {res.res_id: res for res in transformed_resources}
    sorted_resources = [sorting_lookup[res_id] for res_id in upload_order]
    stash = stash_circular_references(sorted_resources, stash_lookup)
    return sorted_resources, stash


def _get_transformed_resources(
    resources: list[XMLResource], intermediary_lookups: IntermediaryLookups
) -> list[IntermediaryResource]:
    result = transform_xmlresources_into_intermediary_resources(resources, intermediary_lookups)
    if result.resource_failures:
        failures = [f"Resource ID: '{x.resource_id}', Message: {x.failure_msg}" for x in result.resource_failures]
        msg = (
            f"{datetime.now()}: WARNING: Unable to create the following resource(s):"
            f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(failures)}"
        )
        raise InputError(msg)
    return result.transformed_resources


def _extract_resources_from_xml(root: etree._Element, default_ontology: str) -> list[XMLResource]:
    resources = list(root.iter(tag="resource"))
    return [XMLResource.from_node(res, default_ontology) for res in resources]



from loguru import logger
from rdflib import Graph


def construct_cardinality_node_shapes(onto: Graph) -> Graph:
    """
    Returns the sh:NodeShapes for the resource classes in the ontology.

    Args:
        onto: ontology graph

    Returns:
        Graph with the resource class node shapes
    """
    logger.info("Constructing cardinality shapes for the ontology.")
    g = Graph()
    g += _construct_resource_nodeshape(onto)
    g += _construct_all_cardinalities(onto)
    return g


def _construct_resource_nodeshape(onto_graph: Graph) -> Graph:
    logger.info("Constructing resource NodeShape")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX dash: <http://datashapes.org/dash#>

    CONSTRUCT {
        ?class a sh:NodeShape ;
                dash:closedByTypes true ;
                sh:property api-shapes:hasPermissions_Cardinality ,
                            [
                              a sh:PropertyShape ;
                              sh:path rdfs:label ;
                              sh:minCount 1 ;
                              sh:maxCount 1 ;
                              sh:severity sh:Violation ;
                              sh:message "A label is required" ;
                            ] ,
                            [ 
                              a sh:PropertyShape ;
                              sh:path knora-api:hasStandoffLinkTo
                            ] .
    } WHERE {
        ?class a owl:Class ;
               knora-api:isResourceClass true ;
               knora-api:canBeInstantiated true .
    }
    """
    if results_graph := onto_graph.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_all_cardinalities(onto_graph: Graph) -> Graph:
    g = Graph()
    g += _construct_1_cardinality(onto_graph)
    g += _construct_0_1_cardinality(onto_graph)
    g += _construct_1_n_cardinality(onto_graph)
    g += _construct_0_n_cardinality(onto_graph)
    return g


def _construct_1_cardinality(onto_graph: Graph) -> Graph:
    logger.info("Constructing cardinality: 1")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX dash: <http://datashapes.org/dash#>

    CONSTRUCT {
      ?class sh:property [
          a sh:PropertyShape ;
          sh:path ?propRestriction ;
          sh:minCount 1 ;
          sh:maxCount 1 ;
          sh:severity sh:Violation ;
          sh:message "Cardinality 1" ;
      ] .
    } WHERE {
      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;
          owl:onProperty ?propRestriction ;
          owl:cardinality 1 .
          
      ?propRestriction knora-api:isEditable true .
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }
    }
    """
    if results_graph := onto_graph.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_0_1_cardinality(onto_graph: Graph) -> Graph:
    logger.info("Constructing cardinality: 0-1")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX dash: <http://datashapes.org/dash#>

    CONSTRUCT {
      ?class sh:property [
          a sh:PropertyShape ;
          sh:path ?propRestriction ;
          sh:minCount 0 ;
          sh:maxCount 1 ;
          sh:severity sh:Violation ;
          sh:message "Cardinality 0-1" ;
      ] .
    } WHERE {
      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;
          owl:onProperty ?propRestriction ;
          owl:maxCardinality 1 .
          
      ?propRestriction knora-api:isEditable true .
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }
    }
    """
    if results_graph := onto_graph.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_1_n_cardinality(onto_graph: Graph) -> Graph:
    logger.info("Constructing cardinality: 1-n")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX dash: <http://datashapes.org/dash#>

    CONSTRUCT {
      ?class sh:property [
          a sh:PropertyShape ;
          sh:path ?propRestriction ;
          sh:minCount 1 ;
          sh:severity sh:Violation ;
          sh:message "Cardinality 1-n" ;
      ] .
    } WHERE {
      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;
          owl:onProperty ?propRestriction ;
          owl:minCardinality 1 .
          
      ?propRestriction knora-api:isEditable true .
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }
    }
    """
    if results_graph := onto_graph.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_0_n_cardinality(onto_graph: Graph) -> Graph:
    logger.info("Constructing cardinality: 0-n")
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX dash: <http://datashapes.org/dash#>

    CONSTRUCT {
      ?class sh:property [
          a sh:PropertyShape ;
          sh:path ?propRestriction ;
      ] .
    } WHERE {
      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;
          owl:onProperty ?propRestriction ;
          owl:minCardinality 0 .
          
      ?propRestriction knora-api:isEditable true .
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }
    }
    """
    if results_graph := onto_graph.query(query_s).graph:
        return results_graph
    return Graph()



from __future__ import annotations

import importlib.resources
import warnings
from pathlib import Path

import regex
from loguru import logger
from lxml import etree

from dsp_tools.error.custom_warnings import DspToolsUserInfo
from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.xml_parsing.parse_xml import parse_xml_file
from dsp_tools.utils.xml_parsing.transform import remove_comments_from_element_tree
from dsp_tools.utils.xml_parsing.transform import transform_into_localnames

separator = "\n    "
list_separator = "\n    - "


def validate_xml_with_schema(xml: etree._Element) -> bool:
    """
    Validates an XML element tree against the DSP XSD schema.

    Args:
        xml: the XML element tree to be validated

    Raises:
        InputError: if the XML file is invalid

    Returns:
        True if the XML file is valid
    """
    cleaned = transform_into_localnames(xml)
    cleaned = remove_comments_from_element_tree(cleaned)
    _warn_user_about_tags_in_simpletext(cleaned)
    problem_msg = validate_xml_against_schema(xml)

    if problem_msg:
        logger.error(problem_msg)
        raise InputError(problem_msg)

    return True


def validate_xml_against_schema(data_xml: etree._Element) -> str | None:
    schema_res = importlib.resources.files("dsp_tools").joinpath("resources/schema/data.xsd")
    with schema_res.open(encoding="utf-8") as schema_file:
        xmlschema = etree.XMLSchema(etree.parse(schema_file))
    if not xmlschema.validate(data_xml):
        error_msg = "The XML file cannot be uploaded due to the following validation error(s):"
        for error in xmlschema.error_log:
            error_msg = f"{error_msg}{separator}Line {error.line}: {error.message}"
        error_msg = error_msg.replace("{https://dasch.swiss/schema}", "")
        return error_msg
    return None


def _warn_user_about_tags_in_simpletext(xml_no_namespace: etree._Element) -> None:
    """
    Checks if there are angular brackets in simple text.
    It is possible that the user mistakenly added XML tags into a simple text field.
    But it is also possible that an angular bracket should be displayed.
    So that the user does not insert XML tags mistakenly into simple text fields,
    the user is warned, if there is any present.
    """
    resources_with_potential_xml_tags = []
    for text in xml_no_namespace.findall(path="resource/text-prop/text"):
        regex_finds_tags = bool(regex.search(r'<([a-zA-Z/"]+|[^\s0-9].*[^\s0-9])>', str(text.text)))
        etree_finds_tags = bool(list(text.iterchildren()))
        has_tags = regex_finds_tags or etree_finds_tags
        if text.attrib["encoding"] == "utf8" and has_tags:
            sourceline = f"line {text.sourceline}: " if text.sourceline else " "
            propname = text.getparent().attrib["name"]  # type: ignore[union-attr]
            resname = text.getparent().getparent().attrib["id"]  # type: ignore[union-attr]
            resources_with_potential_xml_tags.append(f"{sourceline}resource '{resname}', property '{propname}'")
    if resources_with_potential_xml_tags:
        err_msg = (
            "Angular brackets in the format of <text> were found in text properties with encoding=utf8.\n"
            "Please note that these will not be recognised as formatting in the text field, "
            "but will be displayed as-is.\n"
            f"The following resources of your XML file contain angular brackets:{list_separator}"
            f"{list_separator.join(resources_with_potential_xml_tags)}"
        )
        warnings.warn(DspToolsUserInfo(err_msg))


def parse_and_validate_xml_file(input_file: Path | str) -> bool:
    """
    Validates an XML file against the DSP XSD schema.

    Args:
        input_file: path to the XML file to be validated, or parsed ElementTree

    Raises:
        InputError: if the XML file is invalid

    Returns:
        True if the XML file is valid
    """
    root = parse_xml_file(input_file)
    data_xml = remove_comments_from_element_tree(root)
    return validate_xml_with_schema(data_xml)



import json
from typing import Any

from pyld import jsonld
from rdflib import Graph


def serialise_jsonld_for_resource(rdf_graph: Graph) -> dict[str, Any]:
    json_graph = _serialise_json(rdf_graph)
    return _frame_graph_for_resource(json_graph)


def _frame_graph_for_resource(serialised_json: list[dict[str, Any]]) -> dict[str, Any]:
    """
    The result of the serialisation is a list.
    Although the API would accept that, our connection has the payload typed as `dict[str, Any]`.
    If the typing of the connection was changed or a different class was used, this step can be removed.

    Args:
        serialised_json: graph serialised as json-ld

    Returns:
        graph in a specific json-ld format
    """
    json_frame: dict[str, Any] = {"http://api.knora.org/ontology/knora-api/v2#attachedToProject": {}}
    framed: dict[str, Any] = jsonld.frame(serialised_json, json_frame)
    return framed


def serialise_jsonld_for_value(rdf_graph: Graph, resource_iri_str: str) -> dict[str, Any]:
    json_graph = _serialise_json(rdf_graph)
    return _frame_graph_for_value(json_graph, resource_iri_str)


def _frame_graph_for_value(serialised_json: list[dict[str, Any]], resource_iri_str: str) -> dict[str, Any]:
    json_frame = {"@id": resource_iri_str}
    framed: dict[str, Any] = jsonld.frame(serialised_json, json_frame)
    return framed


def _serialise_json(rdf_graph: Graph) -> list[dict[str, Any]]:
    graph_bytes = rdf_graph.serialize(format="json-ld", encoding="utf-8")
    json_graph: list[dict[str, Any]] = json.loads(graph_bytes)
    return json_graph



import unittest
from pathlib import Path

import pandas as pd
import pytest
import regex

from dsp_tools import excel2xml
from dsp_tools.error.exceptions import BaseError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)
# ruff: noqa: PT027 (pytest-unittest-raises-assertion) (remove this line when pytest is used instead of unittest)


class TestExcel2xmlLib(unittest.TestCase):
    def tearDown(self) -> None:
        Path("excel2xml-invalid-data.xml").unlink(missing_ok=True)

    def test_make_xsd_id_compatible(self) -> None:
        teststring = "0aüZ/_-äöü1234567890?`^':.;+*ç%&/()=±“#Ç[]|{}≠₂₃āṇśṣr̥ṁñἄ𝝺𝝲𝛆’الشعرُאדםПопрыгуньяşğ"  # noqa: RUF001
        expected_ = "_0a_Z__-___1234567890_____.__________________________r______________________________"

        # test that the results are distinct from each other
        results = {excel2xml.make_xsd_id_compatible(teststring) for _ in range(10)}
        self.assertTrue(len(results) == 10)
        for res in results:
            self.assertTrue(res.startswith(expected_))

        # test that the results are valid xsd:ids
        for result in results:
            self.assertTrue(regex.search(r"^[a-zA-Z_][\w.-]*$", result))

        # test that invalid inputs lead to an error
        self.assertRaises(BaseError, excel2xml.make_xsd_id_compatible, 0)
        self.assertRaises(BaseError, excel2xml.make_xsd_id_compatible, "n/a")
        self.assertRaises(BaseError, excel2xml.make_xsd_id_compatible, None)
        self.assertRaises(BaseError, excel2xml.make_xsd_id_compatible, "")
        self.assertRaises(BaseError, excel2xml.make_xsd_id_compatible, " ")
        self.assertRaises(BaseError, excel2xml.make_xsd_id_compatible, ".")

        # test that the special characters in the "Label" row of excel2xml-testdata-special-characters.xlsx are replaced
        special_characters_df = pd.read_excel("testdata/excel2xml/excel2xml-testdata-special-characters.xlsx")
        root = excel2xml.make_root("00A1", "test")
        root = excel2xml.append_permissions(root)
        for _, row in special_characters_df.iterrows():
            root.append(
                excel2xml.make_resource(
                    label=row["Label"], restype=":xyz", id=excel2xml.make_xsd_id_compatible(row["Label"])
                )
            )
        # schema validation inside the write_xml() checks if the ids of the resources are valid as xsd:ID
        excel2xml.write_xml(root, "special-characters.xml")
        Path("special-characters.xml").unlink()


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import json
from datetime import datetime

from loguru import logger

from dsp_tools.commands.xmlupload.upload_config import DiagnosticsConfig


def write_id2iri_mapping(
    id2iri_mapping: dict[str, str],
    shortcode: str,
    diagnostics: DiagnosticsConfig,
) -> None:
    """Writes the mapping of internal IDs to IRIs to a file."""
    timestamp = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    servername = diagnostics.server_as_foldername
    id2iri_filename = f"id2iri_{shortcode}_{servername}_{timestamp}.json"
    with open(id2iri_filename, "x", encoding="utf-8") as f:
        json.dump(id2iri_mapping, f, ensure_ascii=False, indent=4)
        print(f"{datetime.now()}: The mapping of internal IDs to IRIs was written to {id2iri_filename}")
        logger.info(f"The mapping of internal IDs to IRIs was written to {id2iri_filename}")



from typing import Protocol


class Problem(Protocol):
    """Information about input errors."""

    def execute_error_protocol(self) -> str:
        """
        This function initiates all the steps for successful problem communication with the user.
        """



from dataclasses import dataclass

import requests
from loguru import logger
from requests import ReadTimeout
from requests import Response

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.legal_info_client import LegalInfoClient
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.request_utils import RequestParameters
from dsp_tools.utils.request_utils import log_and_raise_timeouts
from dsp_tools.utils.request_utils import log_request
from dsp_tools.utils.request_utils import log_response

TIMEOUT = 60

HTTP_LACKING_PERMISSIONS = 403


@dataclass
class LegalInfoClientLive(LegalInfoClient):
    server: str
    project_shortcode: str
    authentication_client: AuthenticationClient

    def post_copyright_holders(self, copyright_holders: list[str]) -> None:
        """Send a list of new copyright holders to the API"""
        logger.debug(f"POST {len(copyright_holders)} new copyright holders")
        try:
            response = self._post_and_log_request("copyright-holders", copyright_holders)
        except (TimeoutError, ReadTimeout) as err:
            log_and_raise_timeouts(err)
        if response.ok:
            return
        if response.status_code == HTTP_LACKING_PERMISSIONS:
            raise BadCredentialsError(
                "Only a project or system administrator can create new copyright holders. "
                "Your permissions are insufficient for this action."
            )
        else:
            raise BaseError(
                f"An unexpected response with the status code {response.status_code} was received from the API. "
                f"Please consult 'warnings.log' for details."
            )

    def _post_and_log_request(self, endpoint: str, data: list[str]) -> Response:
        url = f"{self.server}/admin/projects/shortcode/{self.project_shortcode}/legal-info/{endpoint}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.authentication_client.get_token()}",
        }
        params = RequestParameters("POST", url, TIMEOUT, {"data": data}, headers)
        log_request(params)
        response = requests.post(
            url=params.url,
            headers=params.headers,
            data=params.data_serialized,
            timeout=params.timeout,
        )
        log_response(response)
        return response



from copy import deepcopy

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStash
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash


def stash_circular_references(
    resources: list[IntermediaryResource], stash_lookup: dict[str, list[str]]
) -> Stash | None:
    """Stash the values that would create circular references and remove them from the Resources."""
    stashed_link_values: list[LinkValueStashItem] = []
    stashed_standoff_values: list[StandoffStashItem] = []

    if not stash_lookup:
        return None

    for res in resources:
        if res.res_id not in stash_lookup:
            continue
        links, standoff = _process_one_resource(res, stash_lookup)
        stashed_link_values.extend(links)
        stashed_standoff_values.extend(standoff)

    standoff_stash = StandoffStash.make(stashed_standoff_values)
    link_value_stash = LinkValueStash.make(stashed_link_values)
    return Stash.make(standoff_stash=standoff_stash, link_value_stash=link_value_stash)


def _process_one_resource(
    resource: IntermediaryResource,
    stash_lookup: dict[str, list[str]],
) -> tuple[list[LinkValueStashItem], list[StandoffStashItem]]:
    stashed_link_values: list[LinkValueStashItem] = []
    stashed_standoff_values: list[StandoffStashItem] = []

    for val in resource.values.copy():
        if isinstance(val, IntermediaryLink):
            if val.value_uuid not in stash_lookup[resource.res_id]:
                continue
            stashed_link_values.append(LinkValueStashItem(resource.res_id, resource.type_iri, val))
            resource.values.remove(val)
        elif isinstance(val, IntermediaryRichtext):
            if val.value_uuid not in stash_lookup[resource.res_id]:
                continue
            # val.value is a KnoraStandoffXml text with problematic links.
            # stash it, then replace the problematic text with a UUID
            stashed_standoff_values.append(_stash_standoff(val, resource.res_id, resource.type_iri))

    return stashed_link_values, stashed_standoff_values


def _stash_standoff(value: IntermediaryRichtext, res_id: str, res_type: str) -> StandoffStashItem:
    original_value = deepcopy(value)
    # Replace the content with the UUID
    value.value = FormattedTextValue(value.value_uuid)
    # It is not necessary to add the permissions to the StandoffStashItem.
    # Because when no new permissions are given during an update request,
    # the permissions of the previous value are taken.
    return StandoffStashItem(
        res_id=res_id,
        res_type=res_type,
        value=original_value,
    )



import subprocess
import sys

from dsp_tools.utils.ansi_colors import BOLD_GREEN
from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT


def main() -> None:
    """
    Wrapper for markdown-link-validator.
    markdown-link-validator is a tool that checks the links in Markdown files.
    It prints every single link to the console, and returns a non-zero exit code if there are any errors.
    This script converts the output to a user-friendly short error message.
    """
    cli_command = "just check-links"
    if _get_return_code(cli_command) != 0:
        msg = f"markdown-link-validator reported an error. Run this command to see more details: {cli_command}"
        print(f"{BOLD_RED}{msg}{RESET_TO_DEFAULT}")
        sys.exit(1)
    else:
        msg = "markdown-link-validator: No errors found"
        print(f"{BOLD_GREEN}{msg}{RESET_TO_DEFAULT}")


def _get_return_code(call: str) -> int:
    try:
        res = subprocess.run(call.split(), capture_output=True, check=True, encoding="utf-8")
        return res.returncode
    except subprocess.CalledProcessError as e:
        return e.returncode


if __name__ == "__main__":
    main()
